from __future__ import annotations

import enum
import operator as op
import types
from collections import defaultdict
from collections.abc import AsyncGenerator, AsyncIterator, Callable
from contextlib import suppress
from enum import Enum, StrEnum, auto
from functools import cache
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    Literal,
    NewType,
    NotRequired,
    ParamSpec,
    Protocol,
    Self,
    TypeAlias,
    TypedDict,
    TypeVar,
    TypeVarTuple,
    Union,
    runtime_checkable,
)

# Sort separately due to being a private import
from typing import _GenericAlias  # type: ignore[attr-defined]  # isort: skip  # noqa: PLC2701
from typing import _LiteralGenericAlias  # type: ignore[attr-defined]  # isort: skip  # noqa: PLC2701
from typing import _TypedDictMeta  # type: ignore[attr-defined]  # isort: skip  # noqa: PLC2701
from typing import _ProtocolMeta  # type: ignore[attr-defined]  # isort: skip  # noqa: PLC2701
from typing import _eval_type  # type: ignore[attr-defined]  # isort: skip  # noqa: PLC2701

from collections.abc import Iterable

from django.db.models import (
    Expression,
    F,
    Field,
    ForeignKey,
    ForeignObjectRel,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    Model,
    OneToOneField,
    OneToOneRel,
    Q,
    QuerySet,
    Subquery,
)
from django.db.models.query_utils import RegisterLookupMixin
from graphql import (
    ExecutionResult,
    FieldNode,
    FragmentSpreadNode,
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLNamedType,
    GraphQLNullableType,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLScalarType,
    GraphQLUnionType,
    SelectionNode,
    UndefinedType,
)
from graphql.pyutils import AwaitableOrValue

if TYPE_CHECKING:
    from collections.abc import Container
    from http.cookies import SimpleCookie

    from asgiref.typing import ASGISendEvent
    from django.contrib.auth.models import AbstractUser, AnonymousUser, User
    from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel, GenericRelation
    from django.contrib.sessions.backends.base import SessionBase
    from django.core.files.uploadedfile import UploadedFile
    from django.db.models.sql import Query
    from django.http import QueryDict
    from django.http.request import HttpHeaders, MediaType
    from django.http.response import ResponseHeaders
    from django.test.client import Client
    from django.utils.datastructures import MultiValueDict
    from graphql import (
        DirectiveLocation,
        FormattedExecutionResult,
        FragmentDefinitionNode,
        GraphQLArgumentMap,
        GraphQLFormattedError,
        GraphQLOutputType,
        GraphQLSchema,
        OperationDefinitionNode,
    )
    from graphql.pyutils import Path

    from undine import FilterSet, InterfaceType, MutationType, OrderSet, QueryType, UnionType
    from undine.directives import Directive
    from undine.optimizer.optimizer import OptimizationData
    from undine.utils.graphql.websocket import WebSocketRequest

__all__ = [
    "Annotatable",
    "CalculationArgumentParams",
    "ClientMessage",
    "CombinableExpression",
    "CompleteMessage",
    "ConnectionAckMessage",
    "ConnectionDict",
    "ConnectionInitMessage",
    "ConvertionFunc",
    "DirectiveArgumentParams",
    "DirectiveParams",
    "DispatchProtocol",
    "DjangoExpression",
    "DjangoRequestProtocol",
    "DjangoResponseProtocol",
    "DjangoTestClientResponseProtocol",
    "DocstringParserProtocol",
    "EntrypointParams",
    "EntrypointPermFunc",
    "ErrorMessage",
    "ExecutionResultGen",
    "FieldParams",
    "FieldPermFunc",
    "FilterAliasesFunc",
    "FilterParams",
    "FilterSetParams",
    "ForwardField",
    "GQLInfo",
    "GraphQLFilterResolver",
    "InputParams",
    "InputPermFunc",
    "InterfaceFieldParams",
    "InterfaceTypeParams",
    "JsonObject",
    "Lambda",
    "LiteralArg",
    "M2MChangedParams",
    "ManyMatch",
    "ModelField",
    "MutationKind",
    "MutationTypeParams",
    "NextMessage",
    "NodeDict",
    "ObjectSelections",
    "OptimizerFunc",
    "OrderAliasesFunc",
    "OrderParams",
    "OrderSetParams",
    "PageInfoDict",
    "ParametrizedType",
    "PingMessage",
    "PongMessage",
    "PostDeleteParams",
    "PostSaveParams",
    "PreDeleteParams",
    "PreSaveParams",
    "ProtocolType",
    "QueryTypeParams",
    "RelatedField",
    "RequestMethod",
    "ReverseField",
    "RootTypeParams",
    "Selections",
    "Self",
    "ServerMessage",
    "SubscribeMessage",
    "SupportsLookup",
    "TInterfaceType",
    "TModels",
    "TQueryType",
    "TUnionType",
    "ToManyField",
    "ToOneField",
    "UndineErrorCodes",
    "UnionTypeParams",
    "ValidatorFunc",
    "WebSocketConnectionInitHook",
    "WebSocketConnectionPingHook",
    "WebSocketConnectionPongHook",
    "WebSocketProtocol",
    "WebSocketResult",
]

# Common TypeVars

T = TypeVar("T")
P = ParamSpec("P")
T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)

# Misc.

TypedDictType: TypeAlias = _TypedDictMeta
ParametrizedType: TypeAlias = _GenericAlias
LiteralType: TypeAlias = _LiteralGenericAlias
ProtocolType: TypeAlias = _ProtocolMeta
PrefetchHackCacheType: TypeAlias = defaultdict[str, defaultdict[str, set[str]]]
LiteralArg: TypeAlias = str | int | bytes | bool | Enum | None
TypeHint: TypeAlias = type | types.UnionType | types.GenericAlias
JsonObject: TypeAlias = dict[str, Any] | list[dict[str, Any]]
DefaultValueType: TypeAlias = int | float | str | bool | dict | list | UndefinedType | None
WebSocketResult: TypeAlias = AsyncIterator[ExecutionResult] | ExecutionResult
ExecutionResultGen: TypeAlias = AsyncGenerator[ExecutionResult, None]
SortedSequence: TypeAlias = list[T] | tuple[T, ...]

# Bound TypeVars

TModel = TypeVar("TModel", bound=Model)
TUser = TypeVar("TUser", bound="AbstractUser", covariant=True)  # noqa: PLC0105
TTypedDict = TypeVar("TTypedDict", bound=TypedDictType)
GNT = TypeVar("GNT", bound=GraphQLNullableType)
TTypeHint = TypeVar("TTypeHint", bound=TypeHint)
TQueryType = TypeVar("TQueryType", bound="QueryType")
TUnionType = TypeVar("TUnionType", bound="UnionType")
TInterfaceType = TypeVar("TInterfaceType", bound="InterfaceType")
TInterfaceQueryType = TypeVar("TInterfaceQueryType", bound="QueryType | InterfaceType")
TQueryTypes = TypeVarTuple("TQueryTypes")
TModels = TypeVarTuple("TModels")


# Literals

RequestMethod: TypeAlias = Literal["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "TRACE", "HEAD", "WEBSOCKET"]

# NewTypes

Lambda = NewType("Lambda", types.FunctionType)
"""
Type used to register a different implementations for lambda functions
as opposed to a regular function in the FunctionDispatcher.
"""


# Protocols


@runtime_checkable
class DocstringParserProtocol(Protocol):
    @classmethod
    def parse_body(cls, docstring: str) -> str: ...

    @classmethod
    def parse_arg_descriptions(cls, docstring: str) -> dict[str, str]: ...

    @classmethod
    def parse_return_description(cls, docstring: str) -> str: ...

    @classmethod
    def parse_raise_descriptions(cls, docstring: str) -> dict[str, str]: ...

    @classmethod
    def parse_deprecations(cls, docstring: str) -> dict[str, str]: ...


class DjangoExpression(Protocol):
    """Protocol for any expression that can be used in a Django ORM query."""

    def resolve_expression(
        self,
        query: Query,
        allow_joins: bool,  # noqa: FBT001
        reuse: set[str] | None,
        summarize: bool,  # noqa: FBT001
        for_save: bool,  # noqa: FBT001
    ) -> DjangoExpression: ...


class DispatchProtocol(Protocol[T_co]):
    def __call__(self, key: Any, **kwargs: Any) -> T_co: ...


class MutationDataFunc(Protocol):
    def __call__(
        self,
        instance: Model,
        info: GQLInfo,
        input_data: dict[str, Any],
        mutation_type: type[MutationType],
    ) -> None: ...


class MutationDataCoroutine(Protocol):
    async def __call__(
        self,
        instance: Model,
        info: GQLInfo,
        input_data: dict[str, Any],
        mutation_type: type[MutationType],
    ) -> None: ...


class DjangoRequestProtocol(Protocol[TUser]):  # noqa: PLR0904
    """Protocol of a Django 'HttpRequest' object. Abbreviated to the most useful properties."""

    @property
    def GET(self) -> QueryDict:  # noqa: N802
        """A dictionary-like object containing all given HTTP GET parameters."""

    @property
    def POST(self) -> QueryDict:  # noqa: N802
        """A dictionary-like object containing all given HTTP POST parameters."""

    @property
    def COOKIES(self) -> dict[str, str]:  # noqa: N802
        """A dictionary containing all cookies."""

    @property
    def FILES(self) -> MultiValueDict[str, UploadedFile]:  # noqa: N802
        """A dictionary-like object containing all uploaded files."""

    @property
    def META(self) -> dict[str, Any]:  # noqa: N802
        """A dictionary containing all available HTTP headers."""

    @property
    def scheme(self) -> str | None:
        """A string representing the scheme of the request (http or https usually)."""

    @property
    def path(self) -> str:
        """A string representing the full request path, not including the scheme, domain, or query string."""

    @property
    def method(self) -> RequestMethod:
        """A string representing the HTTP method used in the request."""

    @property
    def headers(self) -> HttpHeaders:
        """A case insensitive, dict-like object for accessing headers in the request."""

    @property
    def body(self) -> bytes:
        """The raw HTTP request body as a bytestring."""

    @property
    def encoding(self) -> str | None:
        """A string representing the current encoding used to decode form submission data."""

    @property
    def user(self) -> TUser | AnonymousUser:
        """The user associated with the request."""

    async def auser(self) -> TUser | AnonymousUser:
        """The user associated with the request."""

    @property
    def session(self) -> SessionBase:
        """A readable and writable, dictionary-like object that represents the current session."""

    @property
    def content_type(self) -> str | None:
        """A string representing the MIME type of the request, parsed from the 'CONTENT_TYPE' header."""

    @property
    def content_params(self) -> dict[str, str] | None:
        """A dictionary of key/value parameters included in the 'CONTENT_TYPE' header."""

    @property
    def accepted_types(self) -> list[MediaType]:
        """A list of 'MediaType' objects representing the accepted content types of the request."""

    @property
    def response_content_type(self) -> str:
        """Response content type as determined from 'accepted_types' in relation to the endpoint's supported types."""

    @response_content_type.setter
    def response_content_type(self, value: str) -> None:
        """Set by decorators in 'undine.http.utils'."""

    def is_secure(self) -> bool:
        """A boolean representing whether the request is over HTTPS."""

    def accepts(self, media_type: str) -> bool:
        """Does the client accept a response in the given media type?"""

    def get_host(self) -> str:
        """Return the HTTP host using the environment or request headers."""

    def get_port(self) -> str:
        """Return the port number for the request as a string."""

    def get_full_path(self, force_append_slash: bool = False) -> str:  # noqa: FBT001,FBT002
        """Return the full path for the request."""

    def get_full_path_info(self, force_append_slash: bool = False) -> str:  # noqa: FBT001,FBT002
        """Return the full path info for the request."""

    def build_absolute_uri(self, location: str | None = None) -> str:
        """Build an absolute URI from the location and the variables available in this request."""


class DjangoResponseProtocol(Protocol):
    """Protocol of a Django 'HttpResponse' object. Abbreviated to the most useful properties."""

    @property
    def status_code(self) -> int:
        """The status code of the response."""

    @property
    def content(self) -> bytes:
        """The content of the response."""

    @property
    def text(self) -> str:
        """The text of the response."""

    @property
    def headers(self) -> ResponseHeaders:
        """The headers of the response."""

    @property
    def cookies(self) -> SimpleCookie:
        """The cookies of the response."""

    @property
    def charset(self) -> str:
        """The charset of the response."""

    @property
    def streaming(self) -> bool:
        """Whether the response is a streaming response."""


class DjangoTestClientResponseProtocol(DjangoResponseProtocol, Protocol):
    """Protocol of a Django 'HttpResponse' object for testing. Abbreviated to the most useful properties."""

    @property
    def client(self) -> Client:
        """The test client instance."""

    @property
    def request(self) -> dict[str, Any]:
        """The request environment data."""

    @property
    def templates(self) -> list[str]:
        """The list of templates used to render the response."""

    @property
    def context(self) -> dict[str, Any]:
        """The template context used to render the template."""

    def json(self) -> dict[str, Any]:
        """The JSON content of the response."""


# Enums


class RelationType(enum.Enum):
    REVERSE_ONE_TO_ONE = "REVERSE_ONE_TO_ONE"
    FORWARD_ONE_TO_ONE = "FORWARD_ONE_TO_ONE"
    FORWARD_MANY_TO_ONE = "FORWARD_MANY_TO_ONE"
    REVERSE_ONE_TO_MANY = "REVERSE_ONE_TO_MANY"
    REVERSE_MANY_TO_MANY = "REVERSE_MANY_TO_MANY"
    FORWARD_MANY_TO_MANY = "FORWARD_MANY_TO_MANY"
    GENERIC_ONE_TO_MANY = "GENERIC_ONE_TO_MANY"
    GENERIC_MANY_TO_ONE = "GENERIC_MANY_TO_ONE"

    @classmethod
    def for_related_field(cls, field: RelatedField | GenericField) -> RelationType:
        field_cls = type(field)
        mapping = cls._related_field_to_relation_type_map()

        for cls_ in field_cls.__mro__:
            with suppress(KeyError):
                return mapping[cls_]

        msg = f"Unknown related field: {field} (of type {field_cls})"
        raise ValueError(msg)

    @enum.property
    def is_reverse(self) -> bool:
        return self in {
            RelationType.REVERSE_ONE_TO_ONE,
            RelationType.REVERSE_ONE_TO_MANY,
            RelationType.REVERSE_MANY_TO_MANY,
            RelationType.GENERIC_ONE_TO_MANY,
        }

    @enum.property
    def is_forward(self) -> bool:
        return self in {
            RelationType.FORWARD_ONE_TO_ONE,
            RelationType.FORWARD_MANY_TO_ONE,
            RelationType.FORWARD_MANY_TO_MANY,
            RelationType.GENERIC_MANY_TO_ONE,
        }

    @enum.property
    def is_generic_relation(self) -> bool:
        return self == RelationType.GENERIC_ONE_TO_MANY

    @enum.property
    def is_generic_foreign_key(self) -> bool:
        return self == RelationType.GENERIC_MANY_TO_ONE

    @enum.property
    def is_single(self) -> bool:
        return self in {
            RelationType.FORWARD_ONE_TO_ONE,
            RelationType.FORWARD_MANY_TO_ONE,
            RelationType.REVERSE_ONE_TO_ONE,
            RelationType.GENERIC_MANY_TO_ONE,
        }

    @enum.property
    def is_many(self) -> bool:
        return self in {
            RelationType.FORWARD_MANY_TO_MANY,
            RelationType.REVERSE_MANY_TO_MANY,
            RelationType.REVERSE_ONE_TO_MANY,
            RelationType.GENERIC_ONE_TO_MANY,
        }

    @enum.property
    def is_many_to_many(self) -> bool:
        return self in {
            RelationType.FORWARD_MANY_TO_MANY,
            RelationType.REVERSE_MANY_TO_MANY,
        }

    @classmethod
    @cache
    def _related_field_to_relation_type_map(cls) -> dict[type[RelatedField | GenericField], RelationType]:
        # Must defer creating this map, since the 'contenttypes' app needs to be loaded first.
        from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation  # noqa: PLC0415

        return {
            OneToOneRel: RelationType.REVERSE_ONE_TO_ONE,  # e.g. Reverse OneToOneField
            ManyToOneRel: RelationType.REVERSE_ONE_TO_MANY,
            ManyToManyRel: RelationType.REVERSE_MANY_TO_MANY,  # e.g. Reverse ManyToManyField
            OneToOneField: RelationType.FORWARD_ONE_TO_ONE,
            ForeignKey: RelationType.FORWARD_MANY_TO_ONE,
            ManyToManyField: RelationType.FORWARD_MANY_TO_MANY,
            GenericRelation: RelationType.GENERIC_ONE_TO_MANY,
            GenericForeignKey: RelationType.GENERIC_MANY_TO_ONE,
        }


class MutationKind(enum.StrEnum):
    create = "create"
    update = "update"
    delete = "delete"
    related = "related"
    custom = "custom"

    @enum.property
    def requires_pk(self) -> bool:
        return self in {MutationKind.update, MutationKind.delete}

    @enum.property
    def no_pk(self) -> bool:
        return self in {MutationKind.create, MutationKind.custom}

    @enum.property
    def should_use_autogeneration(self) -> bool:
        return self in {MutationKind.create, MutationKind.update, MutationKind.related}

    @enum.property
    def should_include_default_value(self) -> bool:
        return self == MutationKind.create

    @enum.property
    def all_inputs_used_by_default(self) -> bool:
        return self == MutationKind.custom


class RelatedAction(enum.StrEnum):
    """What happens to related objects that are not included in mutation input?"""

    null = "null"
    """
    The relation from the related object to the parent object is set to null.
    This will raise an error if the relation is not nullable.
    """

    delete = "delete"
    """
    The related objects are deleted.
    """

    ignore = "ignore"
    """
    The related objects are left unchanged.
    For reverse one-to-one relations, this will raise an error due to the one-to-one constraint.
    """


class ManyMatch(enum.StrEnum):
    any = "any"
    all = "all"
    one_of = "one_of"

    @enum.property
    def operator(self) -> Callable[..., Any]:
        match self:
            case ManyMatch.any:
                return op.or_
            case ManyMatch.all:
                return op.and_
            case ManyMatch.one_of:
                return op.xor
            case _:  # pragma: no cover
                msg = f"Unknown operator '{self}'"
                raise ValueError(msg)


# noinspection PyEnum
class UndineErrorCodes(StrEnum):
    """Error codes for Undine errors."""

    @staticmethod
    def _generate_next_value_(name: str, start: Any, count: int, last_values: list[Any]) -> Any:  # noqa: ARG004
        return name

    ASYNC_ATOMIC_MUTATION_NOT_SUPPORTED = auto()
    ASYNC_NOT_SUPPORTED = auto()
    CONTENT_TYPE_MISSING = auto()
    DATA_LOADER_DID_NOT_RETURN_SORTED_SEQUENCE = auto()
    DATA_LOADER_PRIMING_ERROR = auto()
    DATA_LOADER_WRONG_NUMBER_OF_VALUES_RETURNED = auto()
    DUPLICATE_PRIMARY_KEYS = auto()
    DUPLICATE_TYPE = auto()
    FIELD_NOT_NULLABLE = auto()
    FIELD_ONE_TO_ONE_CONSTRAINT_VIOLATION = auto()
    FILE_NOT_FOUND = auto()
    INVALID_INPUT_DATA = auto()
    INVALID_OPERATION_FOR_METHOD = auto()
    INVALID_ORDER_DATA = auto()
    INVALID_PAGINATION_ARGUMENTS = auto()
    LOOKUP_VALUE_MISSING = auto()
    MISSING_CALCULATION_ARGUMENT = auto()
    MISSING_FILE_MAP = auto()
    MISSING_GRAPHQL_DOCUMENT_PARAMETER = auto()
    MISSING_GRAPHQL_QUERY_AND_DOCUMENT_PARAMETERS = auto()
    MISSING_GRAPHQL_QUERY_PARAMETER = auto()
    MISSING_INSTANCES_TO_DELETE = auto()
    MISSING_OPERATION_NAME = auto()
    MISSING_OPERATIONS = auto()
    MISSING_SUBSCRIPTION_ARGUMENT = auto()
    MODEL_CONSTRAINT_VIOLATION = auto()
    MODEL_INSTANCE_NOT_FOUND = auto()
    MUTATION_TOO_MANY_OBJECTS = auto()
    MUTATION_TREE_MODEL_MISMATCH = auto()
    NO_EVENT_STREAM = auto()
    NO_EXECUTION_RESULT = auto()
    NO_OPERATION = auto()
    NODE_ID_NOT_GLOBAL_ID = auto()
    NODE_INTERFACE_MISSING = auto()
    NODE_INVALID_GLOBAL_ID = auto()
    NODE_MISSING_OBJECT_TYPE = auto()
    NODE_QUERY_TYPE_ID_FIELD_MISSING = auto()
    NODE_QUERY_TYPE_MISSING = auto()
    NODE_TYPE_NOT_OBJECT_TYPE = auto()
    OPERATION_NOT_FOUND = auto()
    OPTIMIZER_ERROR = auto()
    PERMISSION_DENIED = auto()
    PERSISTED_DOCUMENT_NOT_FOUND = auto()
    PERSISTED_DOCUMENTS_NOT_SUPPORTED = auto()
    PRIMARY_KEYS_MISSING = auto()
    RELATION_NOT_NULLABLE = auto()
    REQUEST_DECODING_ERROR = auto()
    REQUEST_PARSE_ERROR = auto()
    SCALAR_CONVERSION_ERROR = auto()
    SCALAR_INVALID_VALUE = auto()
    SCALAR_TYPE_NOT_SUPPORTED = auto()
    SUBSCRIPTION_TIMEOUT = auto()
    TOO_MANY_FILTERS = auto()
    TOO_MANY_ORDERS = auto()
    TYPED_DICT_ANNOTATED_INCORRECT_METADATA = auto()
    UNEXPECTED_CALCULATION_ARGUMENT = auto()
    UNEXPECTED_ERROR = auto()
    UNEXPECTED_SUBSCRIPTION_ARGUMENT = auto()
    UNION_RESOLVE_TYPE_INVALID_VALUE = auto()
    UNION_RESOLVE_TYPE_MODEL_NOT_FOUND = auto()
    UNSUPPORTED_CONTENT_TYPE = auto()
    USE_WEBSOCKETS_FOR_SUBSCRIPTIONS = auto()
    VALIDATION_ABORTED = auto()
    VALIDATION_ERROR = auto()


# Model

ToOneField: TypeAlias = OneToOneField | OneToOneRel | ForeignKey
ToManyField: TypeAlias = ManyToManyField | ManyToManyRel | ManyToOneRel
ForwardField: TypeAlias = OneToOneField | ForeignKey | ManyToManyField
ReverseField: TypeAlias = OneToOneRel | ManyToManyRel | ManyToOneRel
RelatedField: TypeAlias = ToOneField | ToManyField
GenericField: TypeAlias = Union["GenericForeignKey", "GenericRelation", "GenericRel"]
ModelField: TypeAlias = Field | ForeignObjectRel
CombinableExpression: TypeAlias = Expression | Subquery
Annotatable: TypeAlias = CombinableExpression | F | Q
SupportsLookup: TypeAlias = RegisterLookupMixin | type[RegisterLookupMixin]
QuerySetMap: TypeAlias = dict[type["QueryType"], QuerySet]

# GraphQL


class GQLInfo(GraphQLResolveInfo, Generic[TUser]):
    """GraphQL execution information given to a GraphQL field resolver."""

    field_name: str
    """Name of the field being resolved."""

    field_nodes: list[FieldNode]
    """
    GraphQL AST Field Nodes in the GraphQL operation for which this field is being resolved for.
    If the same field is queried with a different alias, it will be resolved separately.
    """

    return_type: GraphQLOutputType
    """The GraphQL type of the resolved field."""

    parent_type: GraphQLObjectType
    """The GraphQL type to which this field belongs."""

    path: Path
    """
    Path from the root field to the current field.
    Last part is the field's alias, if one is given, otherwise it's the field's name.
    """

    schema: GraphQLSchema
    """The schema where the GraphQL operation is being executed."""

    fragments: dict[str, FragmentDefinitionNode]
    """A dictionary of GraphQL AST Fragment Definition Nodes in the GraphQL Document."""

    root_value: Any
    """GraphQL root value. Set by `undine_settings.ROOT_VALUE`."""

    operation: OperationDefinitionNode
    """The GraphQL AST Operation Definition Node currently being executed."""

    variable_values: dict[str, Any]
    """The variables passed to the GraphQL operation."""

    context: DjangoRequestProtocol[TUser]
    """The context passed to the GraphQL operation. This is always the Django request object."""

    is_awaitable: Callable[[Any], bool]
    """Function for testing whether the GraphQL resolver is awaitable or not."""


UniquelyNamedGraphQLElement: TypeAlias = (
    GraphQLScalarType
    | GraphQLObjectType
    | GraphQLInterfaceType
    | GraphQLUnionType
    | GraphQLEnumType
    | GraphQLInputObjectType
    | GraphQLDirective
)
HasGraphQLExtensions: TypeAlias = (
    GraphQLNamedType | GraphQLDirective | GraphQLField | GraphQLInputField | GraphQLArgument | GraphQLEnumValue
)


Selections: TypeAlias = Iterable[SelectionNode]
ObjectSelections: TypeAlias = Iterable[FieldNode | FragmentSpreadNode]


class NodeDict(TypedDict, Generic[TModel]):
    cursor: str
    node: TModel


class PageInfoDict(TypedDict):
    hasNextPage: bool
    hasPreviousPage: bool
    startCursor: str | None
    endCursor: str | None


class ConnectionDict(TypedDict, Generic[TModel]):
    totalCount: int
    pageInfo: PageInfoDict
    edges: list[NodeDict[TModel]]


# TypedDicts


class RootTypeParams(TypedDict, total=False):
    """Arguments for an Undine `RootType`."""

    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class EntrypointParams(TypedDict, total=False):
    """Arguments for an Undine `Entrypoint`."""

    many: bool
    nullable: bool
    limit: int
    description: str | None
    deprecation_reason: str | None
    complexity: int
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class QueryTypeParams(TypedDict, total=False):
    """Arguments for an Undine `QueryType`."""

    model: type[Model]
    filterset: type[FilterSet]
    orderset: type[OrderSet]
    auto: bool
    exclude: list[str]
    interfaces: list[type[InterfaceType]]
    register: bool
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class FieldParams(TypedDict, total=False):
    """Arguments for an Undine `Field`."""

    many: bool
    nullable: bool
    description: str | None
    deprecation_reason: str | None
    complexity: int
    field_name: str
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class InterfaceTypeParams(TypedDict, total=False):
    """Arguments for an Undine `InterfaceType`."""

    interfaces: list[type[InterfaceType]]
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class InterfaceFieldParams(TypedDict, total=False):
    """Arguments for an Undine `InterfaceField`."""

    args: GraphQLArgumentMap
    description: str | None
    deprecation_reason: str | None
    field_name: str
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class UnionTypeParams(TypedDict, total=False):
    """Arguments for an Undine `UnionType`."""

    schema_name: str
    filterset: type[FilterSet]
    orderset: type[OrderSet]
    directives: list[Directive]
    extensions: dict[str, Any]


class MutationTypeParams(TypedDict, total=False):
    """Arguments for an Undine `MutationType`."""

    model: type[Model]
    kind: Literal["create", "update", "delete", "related", "custom"]
    related_action: Literal["null", "delete", "ignore"]
    auto: bool
    exclude: list[str]
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class InputParams(TypedDict, total=False):
    """Arguments for an Undine `Input`."""

    many: bool
    required: bool
    default_value: DefaultValueType
    input_only: bool
    hidden: bool
    description: str | None
    deprecation_reason: str | None
    field_name: str
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class FilterSetParams(TypedDict, total=False):
    """Arguments for an Undine `FilterSet`."""

    auto: bool
    exclude: list[str]
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class FilterParams(TypedDict, total=False):
    """Arguments for an Undine `Filter`."""

    lookup: str
    many: bool
    match: Literal["any", "all", "one_of"]
    distinct: bool
    required: bool
    empty_values: Container[Any]
    description: str | None
    deprecation_reason: str | None
    field_name: str
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class OrderSetParams(TypedDict, total=False):
    """Arguments for an Undine `OrderSet`."""

    model: type[Model]
    auto: bool
    exclude: list[str]
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class OrderParams(TypedDict, total=False):
    """Arguments for an Undine `Order`."""

    null_placement: Literal["first", "last"]
    description: str | None
    deprecation_reason: str | None
    field_name: str
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class DirectiveParams(TypedDict, total=False):
    """Arguments for an Undine `Directive`."""

    locations: list[DirectiveLocation | str]
    is_repeatable: bool
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class DirectiveArgumentParams(TypedDict, total=False):
    """Arguments for an Undine `DirectiveArgument`."""

    default_value: DefaultValueType
    description: str | None
    deprecation_reason: str | None
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class CalculationArgumentParams(TypedDict, total=False):
    """Arguments for an Undine `DirectiveArgument`."""

    default_value: DefaultValueType
    description: str | None
    deprecation_reason: str | None
    schema_name: str
    directives: list[Directive]
    extensions: dict[str, Any]


class PreSaveParams(TypedDict, Generic[TModel]):
    """Parameters for a pre-save signal"""

    sender: type[TModel]
    """The model whose instance is being saved"""

    instance: TModel
    """The instance being saved"""

    raw: bool
    """
    Is the model is saved exactly as presented (i.e. when loading a fixture).
    One should not query/modify other records in the database as the database might not be in a consistent state yet.
    """

    using: str
    """The database alias being used"""

    update_fields: set[str] | None
    """The fields that are being updated (as passed to Model.save())"""


class PostSaveParams(TypedDict, Generic[TModel]):
    """Parameters for a post-save signal"""

    sender: type[TModel]
    """The model whose instance was saved"""

    instance: TModel
    """The instance was saved"""

    created: bool
    """Whether the instance was created or updated"""

    raw: bool
    """
    Is the model is saved exactly as presented (i.e. when loading a fixture).
    One should not query/modify other records in the database as the database might not be in a consistent state yet.
    """

    using: str
    """The database alias being used"""

    update_fields: set[str] | None
    """The fields that are being updated (as passed to Model.save())"""


class PreDeleteParams(TypedDict, Generic[TModel]):
    """Parameters for a pre-delete signal"""

    sender: type[TModel]
    """The model whose instance is being deleted"""

    instance: TModel
    """The instance being deleted"""

    using: str
    """The database alias being used"""

    origin: TModel | QuerySet[TModel]
    """The Model or QuerySet instance from which the deletion originated."""


class PostDeleteParams(TypedDict, Generic[TModel]):
    """Parameters for a post-delete signal"""

    sender: type[TModel]
    """The model whose instance was deleted"""

    instance: TModel
    """The instance that was deleted.
    Note that the instance will no longer be in the database,
    so its pk will be None and all relations have been disconnected.
    """

    using: str
    """The database alias being used"""

    origin: TModel | QuerySet[TModel]
    """The Model or QuerySet instance from which the deletion originated."""


class M2MChangedParams(TypedDict):
    """Parameters for a m2m-changed signal"""

    sender: type[Model]
    """The through model that is being modified"""

    instance: Model
    """
    The instance whose many to many relation is being modified.
    Can be either the model with the many to many relation, or the related model.
    The 'reverse' argument will be `True` if it's the latter.
    """

    action: Literal["pre_add", "post_add", "pre_remove", "post_remove"]
    """The action being performed on the relation"""

    reverse: bool
    """Whether the reverse relation being modified or not"""

    model: type[Model]
    """
    The model whose instances are being added to or removed from the relation.
    Can be either the model with the many to many relation, or the related model.
    The 'reverse' argument will be `False` if it's the latter.
    """

    pk_set: set[Any]
    """The primary keys of the instances being added to or removed from the relation"""

    using: str
    """The database alias being used"""


class PostgresFTSLangSpecificFields(TypedDict, total=False):
    """Specify fields for text search per language."""

    # Cannot use "Iterable[str]" since "str" is also an iterable of strings.
    arabic: list[str] | tuple[str, ...] | set[str]
    armenian: list[str] | tuple[str, ...] | set[str]
    basque: list[str] | tuple[str, ...] | set[str]
    catalan: list[str] | tuple[str, ...] | set[str]
    danish: list[str] | tuple[str, ...] | set[str]
    dutch: list[str] | tuple[str, ...] | set[str]
    english: list[str] | tuple[str, ...] | set[str]
    finnish: list[str] | tuple[str, ...] | set[str]
    french: list[str] | tuple[str, ...] | set[str]
    german: list[str] | tuple[str, ...] | set[str]
    greek: list[str] | tuple[str, ...] | set[str]
    hindi: list[str] | tuple[str, ...] | set[str]
    hungarian: list[str] | tuple[str, ...] | set[str]
    indonesian: list[str] | tuple[str, ...] | set[str]
    irish: list[str] | tuple[str, ...] | set[str]
    italian: list[str] | tuple[str, ...] | set[str]
    lithuanian: list[str] | tuple[str, ...] | set[str]
    nepali: list[str] | tuple[str, ...] | set[str]
    norwegian: list[str] | tuple[str, ...] | set[str]
    portuguese: list[str] | tuple[str, ...] | set[str]
    romanian: list[str] | tuple[str, ...] | set[str]
    russian: list[str] | tuple[str, ...] | set[str]
    serbian: list[str] | tuple[str, ...] | set[str]
    spanish: list[str] | tuple[str, ...] | set[str]
    swedish: list[str] | tuple[str, ...] | set[str]
    tamil: list[str] | tuple[str, ...] | set[str]
    turkish: list[str] | tuple[str, ...] | set[str]
    yiddish: list[str] | tuple[str, ...] | set[str]


FTSLang: TypeAlias = Literal[
    "arabic",
    "armenian",
    "basque",
    "catalan",
    "danish",
    "dutch",
    "english",
    "finnish",
    "french",
    "german",
    "greek",
    "hindi",
    "hungarian",
    "indonesian",
    "irish",
    "italian",
    "lithuanian",
    "nepali",
    "norwegian",
    "portuguese",
    "romanian",
    "russian",
    "serbian",
    "spanish",
    "swedish",
    "tamil",
    "turkish",
    "yiddish",
]

LangCode: TypeAlias = Literal[
    "ar",
    "hy",
    "eu",
    "ca",
    "da",
    "nl",
    "en",
    "fi",
    "fr",
    "de",
    "el",
    "hi",
    "hu",
    "id",
    "ga",
    "it",
    "lt",
    "ne",
    "nb",
    "pt",
    "ro",
    "ru",
    "sr",
    "es",
    "sv",
    "ta",
    "tr",
    "yi",
]

LangSep: TypeAlias = Literal["|", "&", "<->", "<1>", "<2>", "<3>", "<4>", "<5>", "<6>", "<7>", "<8>", "<9>"]


# Resolvers

_AnyValue: TypeAlias = Annotated[Any, "value"]
_AnyModel: TypeAlias = Annotated[Any, "django.db.models.Model"]
_AnyField: TypeAlias = Annotated[Any, "undine.Field"]
_AnyInput: TypeAlias = Annotated[Any, "undine.Input"]
_AnyFilter: TypeAlias = Annotated[Any, "undine.Filter"]
_AnyOrder: TypeAlias = Annotated[Any, "undine.Order"]

EntrypointPermFunc: TypeAlias = Callable[[Any, GQLInfo, _AnyValue], AwaitableOrValue[None]]
FieldPermFunc: TypeAlias = Callable[[_AnyModel, GQLInfo, _AnyValue], AwaitableOrValue[None]]
InputPermFunc: TypeAlias = Callable[[_AnyModel, GQLInfo, _AnyValue], AwaitableOrValue[None]]
ValidatorFunc: TypeAlias = Callable[[_AnyModel, GQLInfo, _AnyValue], AwaitableOrValue[None]]

ConvertionFunc: TypeAlias = Callable[[_AnyInput, _AnyValue], _AnyValue]
VisibilityFunc: TypeAlias = Callable[[Any, DjangoRequestProtocol], bool]

OptimizerFunc: TypeAlias = Callable[[_AnyField, "OptimizationData", GQLInfo], None]


class FilterAliasesFunc(Protocol):
    def __call__(self, root: _AnyFilter, /, info: GQLInfo, *, value: Any) -> dict[str, DjangoExpression]: ...


class OrderAliasesFunc(Protocol):
    def __call__(self, root: _AnyOrder, /, info: GQLInfo, *, descending: bool) -> dict[str, DjangoExpression]: ...


class GraphQLFilterResolver(Protocol):
    def __call__(self, root: _AnyFilter, /, info: GQLInfo, *, value: Any) -> Q: ...


# Callbacks

QuerySetCallback: TypeAlias = Callable[[GQLInfo], QuerySet]
FilterCallback: TypeAlias = Callable[[QuerySet, GQLInfo], QuerySet]
PersistedDocumentsPermissionsCallback: TypeAlias = Callable[[DjangoRequestProtocol, dict[str, str]], None]


def eval_type(type_: Any, *, globals_: dict[str, Any] | None = None, locals_: dict[str, Any] | None = None) -> Any:
    """
    Evaluate a type, possibly using the given globals and locals.

    This is a proxy of the 'typing._eval_type' function.
    """
    return _eval_type(type_, globals_ or {}, locals_ or {})  # pragma: no cover


# Subscriptions


class ConnectionInitMessage(TypedDict):
    type: Literal["connection_init"]
    payload: NotRequired[dict[str, Any] | None]


class ConnectionAckMessage(TypedDict):
    type: Literal["connection_ack"]
    payload: NotRequired[dict[str, Any] | None]


class PingMessage(TypedDict):
    type: Literal["ping"]
    payload: NotRequired[dict[str, Any] | None]


class PongMessage(TypedDict):
    type: Literal["pong"]
    payload: NotRequired[dict[str, Any] | None]


class SubscribeMessage(TypedDict):
    type: Literal["subscribe"]
    id: str
    payload: dict[str, Any]  # GraphQL operation params


class NextMessage(TypedDict):
    type: Literal["next"]
    id: str
    payload: FormattedExecutionResult


class ErrorMessage(TypedDict):
    type: Literal["error"]
    id: str
    payload: list[GraphQLFormattedError]


class CompleteMessage(TypedDict):
    type: Literal["complete"]
    id: str


ClientMessage: TypeAlias = ConnectionInitMessage | PingMessage | PongMessage | SubscribeMessage | CompleteMessage
"""Messages sent by the client."""

ServerMessage: TypeAlias = (
    ConnectionAckMessage | PingMessage | PongMessage | NextMessage | ErrorMessage | CompleteMessage
)
"""Messages sent by the server."""


class UrlRoute(TypedDict):
    args: tuple[str, ...]
    kwargs: dict[str, str]


class WebSocketASGIScope(TypedDict):
    type: str
    asgi: dict[str, str]
    http_version: str
    scheme: str
    server: tuple[str, int]
    client: tuple[str, int]
    root_path: str
    path: str
    raw_path: bytes
    query_string: bytes
    headers: list[tuple[bytes, bytes]]
    subprotocols: list[str]
    state: dict[str, Any]
    extensions: dict[str, Any]
    cookies: dict[str, str]
    session: SessionBase
    user: User | AnonymousUser
    path_remaining: str
    url_route: UrlRoute


class GraphQLWebSocketCloseCode(enum.IntEnum):
    """Possible WebSocket close codes."""

    # WebSocket Protocol.
    # See: https://datatracker.ietf.org/doc/html/rfc6455#section-7.4.1

    NORMAL_CLOSURE = 1000
    """Client has closed the connection normally."""

    GOING_AWAY = 1001
    """Browser tab closing, graceful server shutdown."""

    PROTOCOL_ERROR = 1002
    """Endpoint received malformed frame."""

    UNSUPPORTED_DATA = 1003
    """Endpoint received unsupported frame."""

    NO_STATUS_RCVD = 1005
    """Got no close status but transport layer finished normally."""

    ABNORMAL_CLOSURE = 1006
    """Transport layer broke."""

    INVALID_FRAME_PAYLOAD_DATA = 1007
    """Data in endpoint's frame is not consistent."""

    POLICY_VIOLATION = 1008
    """Generic code not applicable to any other."""

    MESSAGE_TOO_BIG = 1009
    """Endpoint won't process large message."""

    MANDATORY_EXTENSION = 1010
    """Client wanted extension(s) that server did not negotiate."""

    INTERNAL_ERROR = 1011
    """Unexpected server problem while operating."""

    SERVICE_RESTART = 1012
    """Server/service is restarting."""

    TRY_AGAIN_LATER = 1013
    """Temporary server condition forced blocking client's application-based request."""

    BAD_GATEWAY = 1014
    """Server acting as gateway/proxy got invalid response."""

    TLS_HANDSHAKE = 1015
    """Transport layer broke because TLS handshake failed."""

    # GraphQL over WebSocket Protocol.
    # See: https://github.com/graphql/graphql-over-http/blob/main/rfcs/GraphQLOverWebSocket.md

    BAD_REQUEST = 4400
    """Client has sent and invalid message to the Server."""

    UNAUTHORIZED = 4401
    """Client has not received a ConnectionAck response before attempting to sent a Subscribe message."""

    FORBIDDEN = 4403
    """Server has rejected the connection init attempt."""

    BAD_RESPONSE = 4004
    """Client received an invalid message from the Server."""

    INTERNAL_CLIENT_ERROR = 4005
    """An unexpected error occurred on the Client."""

    SUBPROTOCOL_NOT_ACCEPTABLE = 4406
    """Server does not support the websocket sub-protocol requested by the Client."""

    CONNECTION_INITIALISATION_TIMEOUT = 4408
    """Server did not receive a valid ConnectionInit message from the Client in the specified time."""

    SUBSCRIBER_ALREADY_EXISTS = 4409
    """A Subscribe operation already exists for the ID given by the Client."""

    TOO_MANY_INITIALISATION_REQUESTS = 4429
    """More than one ConnectionInit request made by the Client."""

    INTERNAL_SERVER_ERROR = 4500
    """An unexpected error occurred on the Server."""

    CONNECTION_ACKNOWLEDGEMENT_TIMEOUT = 4504
    """Server did not send the ConnectionAck message to the Client in the specified time."""


WebSocketConnectionInitHook: TypeAlias = Callable[["WebSocketRequest"], AwaitableOrValue[dict[str, Any] | None]]
WebSocketConnectionPingHook: TypeAlias = Callable[["WebSocketRequest"], AwaitableOrValue[dict[str, Any] | None]]
WebSocketConnectionPongHook: TypeAlias = Callable[["WebSocketRequest"], AwaitableOrValue[None]]


class WebSocketProtocol(Protocol):
    @property
    def scope(self) -> WebSocketASGIScope: ...

    async def send(self, message: ASGISendEvent) -> None: ...



from __future__ import annotations

from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.db.models import Field, ForeignKey, Model, OneToOneField, OneToOneRel
from graphql import Undefined

from undine import Input, MutationType
from undine.converters import convert_to_default_value
from undine.dataclasses import LazyLambda, TypeRef
from undine.exceptions import ModelFieldDoesNotExistError
from undine.typing import ToManyField
from undine.utils.model_utils import get_model_field


@convert_to_default_value.register
def _(ref: Field, **kwargs: Any) -> Any:
    if ref.has_default() and not callable(ref.default):
        return ref.default
    if ref.null:
        return None
    if ref.blank and ref.empty_strings_allowed:
        return ""
    return Undefined


@convert_to_default_value.register
def _(ref: OneToOneField | ForeignKey, **kwargs: Any) -> Any:
    if ref.null:
        return None
    return Undefined


@convert_to_default_value.register
def _(_: OneToOneRel | ToManyField, **kwargs: Any) -> Any:
    return Undefined


@convert_to_default_value.register
def _(_: GenericRelation, **kwargs: Any) -> Any:
    return Undefined


@convert_to_default_value.register
def _(_: GenericForeignKey, **kwargs: Any) -> Any:
    return Undefined


@convert_to_default_value.register
def _(_: TypeRef, **kwargs: Any) -> Any:
    return Undefined


@convert_to_default_value.register
def _(_: LazyLambda, **kwargs: Any) -> Any:
    return Undefined


@convert_to_default_value.register
def _(_: FunctionType, **kwargs: Any) -> Any:
    return Undefined


@convert_to_default_value.register
def _(_: type[MutationType], **kwargs: Any) -> Any:
    return Undefined


@convert_to_default_value.register
def _(_: type[Model], **kwargs: Any) -> Any:
    caller: Input = kwargs["caller"]

    try:
        field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    except ModelFieldDoesNotExistError:
        return Undefined

    return convert_to_default_value(field, **kwargs)



from __future__ import annotations

from inspect import cleandoc
from typing import TYPE_CHECKING, Any

from undine.settings import undine_settings

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from types import FunctionType

__all__ = [
    "comma_sep_str",
    "dotpath",
    "get_docstring",
    "to_camel_case",
    "to_pascal_case",
    "to_schema_name",
    "to_snake_case",
]


def to_camel_case(string: str) -> str:
    """Convert string from `snake_case` or `PascalCase` to `camelCase`."""
    if not string:
        return ""

    words = string.split("_")

    # No underscores, should be camelCase or PascalCase
    if len(words) == 1:
        return string[0].lower() + string[1:]

    text = words[0]
    for word in words[1:]:
        text += word.capitalize()
    return text


def to_pascal_case(string: str) -> str:
    """Convert string from `snake_case` or `camelCase` to `PascalCase`."""
    if not string:
        return ""

    words = string.split("_")

    # No underscores, should be camelCase or PascalCase
    if len(words) == 1:
        return string[0].upper() + string[1:]

    text = ""
    for word in words:
        text += word.capitalize()
    return text


def to_snake_case(string: str) -> str:
    """Convert string from `camelCase` or `PascalCase` to `snake_case`."""
    text: str = ""
    for char in string:
        if char.isupper() and text:
            text += "_"
        text += char.lower()
    return text


def dotpath(obj: type | FunctionType | Callable[..., Any]) -> str:
    """Get the 'dotpath import path' of the given object."""
    return f"{obj.__module__}.{obj.__qualname__}"


def to_schema_name(name: str) -> str:
    """Convert the given name to camelCase if using camelCased schema names."""
    if undine_settings.CAMEL_CASE_SCHEMA_FIELDS:
        return to_camel_case(name)
    return name  # pragma: no cover


def get_docstring(ref: Any) -> str | None:
    """Get the docstring of the given object, if it has one."""
    docstring = getattr(ref, "__doc__", None)
    if docstring is None:
        return None
    return cleandoc(docstring).strip() or None


def comma_sep_str(values: Iterable[Any], *, last_sep: str = "&", quote: bool = False) -> str:
    """
    Return a comma separated string of the given values,
    with the value of `last_sep` before the last value.
    Remove any empty values.

    >>> comma_sep_str(["foo", "bar", "baz"])
    "foo, bar & baz"

    >>> comma_sep_str(["foo", "bar", "baz"], last_sep="or", quote=True)
    "'foo', 'bar' or 'baz'"
    """
    string: str = ""
    previous_value: str = ""
    values_iter = iter(values)
    try:
        while True:
            value = str(next(values_iter))
            if not value:
                continue
            if previous_value:
                if string:
                    string += ", "
                string += f"'{previous_value}'" if quote else previous_value
            previous_value = value
    except StopIteration:
        if string:
            string += f" {last_sep} "
        string += f"'{previous_value}'" if quote else previous_value

    return string



from __future__ import annotations

from copy import copy
from typing import TYPE_CHECKING

from django.db.models import F, ManyToManyField, ManyToManyRel, OuterRef, Value, Window
from django.db.models.functions import Greatest, RowNumber

from undine import InterfaceType, QueryType, UnionType
from undine.dataclasses import ValidatedPaginationArgs
from undine.exceptions import GraphQLPaginationArgumentValidationError
from undine.optimizer.prefetch_hack import register_for_prefetch_hack
from undine.settings import undine_settings
from undine.utils.model_utils import SubqueryCount
from undine.utils.reflection import is_subclass

if TYPE_CHECKING:
    from django.db.models import QuerySet

    from undine import GQLInfo
    from undine.typing import CombinableExpression, ToManyField


__all__ = [
    "OffsetPagination",
    "PaginationHandler",
]


class PaginationHandler:
    """Handles pagination of a queryset based on the given arguments."""

    def __init__(
        self,
        *,
        typename: str,
        after: str | None = None,
        before: str | None = None,
        first: int | None = None,
        last: int | None = None,
        offset: int | None = None,
        limit: int | None = None,
        page_size: int | None = None,
    ) -> None:
        """
        Create a new PaginationHandler.

        :param typename: The typename of the GraphQL type to paginate.
        :param first: Number of item to return from the start.
        :param last: Number of item to return from the end (after applying `first`).
        :param after: Cursor value for the last item in the previous page.
        :param before: Cursor value for the first item in the next page.
        :param offset: Number of item to skip from the start. No offset if `None`.
        :param limit: Maximum limit for the number of item that can be requested in a page. No limit if `None`.
        :param page_size: Maximum limit for the number of item that can be requested in a page. No limit if `None`.
        """
        validated_args = self.validate(
            typename=typename,
            first=first,
            last=last,
            after_cursor=after,
            before_cursor=before,
            offset=offset,
            limit=limit,
            page_size=page_size,
        )

        self.after = validated_args.after
        """The index after which to start (exclusive)."""

        self.before = validated_args.before
        """The index before which to stop (exclusive)."""

        self.first = validated_args.first
        """The number of items to return from the start."""

        self.last = validated_args.last
        """The number of items to return from the end (after evaluating first)."""

        self.page_size = page_size
        """Maximum number of item to return in a page. No limit if `None`."""

        # Calculated in `paginate_queryset` or `paginate_prefetch_queryset` if needed.
        self.start: int = 0
        """The index to start the pagination from."""

        # Calculated in `paginate_queryset` or `paginate_prefetch_queryset` if needed.
        self.stop: int | None = None
        """The index to stop the pagination at."""

        # Modified in `paginate_queryset` or `paginate_prefetch_queryset` if needed.
        self.total_count: int | None = None
        """The total number of items that can be paginated."""

        # Modified during optimization based on pagination params.
        self.requires_total_count: bool = False
        """Whether the total count is required for this query."""

    @staticmethod
    def validate(  # noqa: C901, PLR0912, PLR0915
        *,
        typename: str,
        #
        # Cursor based pagination
        first: int | None,
        last: int | None,
        after_cursor: str | None,
        before_cursor: str | None,
        #
        # Offset based pagination
        offset: int | None,
        limit: int | None,
        #
        # Page size
        page_size: int | None,
    ) -> ValidatedPaginationArgs:
        """Validate the given pagination arguments and return the validated arguments."""
        after: int | None = None
        before: int | None = None

        if after_cursor is not None:
            from undine.relay import cursor_to_offset  # noqa: PLC0415

            try:
                after = cursor_to_offset(typename, after_cursor)
            except Exception as error:
                msg = f"Argument 'after' is not a valid cursor for type '{typename}'."
                raise GraphQLPaginationArgumentValidationError(msg) from error

        if before_cursor is not None:
            from undine.relay import cursor_to_offset  # noqa: PLC0415

            try:
                before = cursor_to_offset(typename, before_cursor)
            except Exception as error:
                msg = f"Argument 'before' is not a valid cursor for type '{typename}'."
                raise GraphQLPaginationArgumentValidationError(msg) from error

        if page_size is not None and (not isinstance(page_size, int) or page_size < 1):
            msg = f"`page_size` must be `None` or a positive integer, got: {page_size!r}"
            raise GraphQLPaginationArgumentValidationError(msg)

        if first is not None:
            if not isinstance(first, int) or first <= 0:
                msg = "Argument 'first' must be a positive integer."
                raise GraphQLPaginationArgumentValidationError(msg)

            if isinstance(page_size, int) and first > page_size:
                msg = f"Requesting first {first} records exceeds the maximum page size of {page_size}."
                raise GraphQLPaginationArgumentValidationError(msg)

        if last is not None:
            if not isinstance(last, int) or last <= 0:
                msg = "Argument 'last' must be a positive integer."
                raise GraphQLPaginationArgumentValidationError(msg)

            if isinstance(page_size, int) and last > page_size:
                msg = f"Requesting last {last} records exceeds the maximum page size of {page_size}."
                raise GraphQLPaginationArgumentValidationError(msg)

        if offset is not None:
            if after is not None or before is not None:
                msg = "Can only use either `offset` or `before`/`after` for pagination."
                raise GraphQLPaginationArgumentValidationError(msg)

            if not isinstance(offset, int) or offset < 0:
                msg = "Argument `offset` must be a positive integer."
                raise GraphQLPaginationArgumentValidationError(msg)

            # Convert offset to after cursor value. Note that after cursor dictates
            # a value _after_ which results should be returned, so we need to subtract
            # 1 from the offset to get the correct cursor value.
            if offset > 0:  # ignore zero offset
                after = offset - 1

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                msg = "Argument `limit` must be a positive integer."
                raise GraphQLPaginationArgumentValidationError(msg)

            if isinstance(page_size, int) and limit > page_size:
                msg = f"Requesting {limit} records exceeds the maximum page size of {page_size}."
                raise GraphQLPaginationArgumentValidationError(msg)

            # In offset pagination, limit is the same as requesting the first n items
            first = limit

        if isinstance(page_size, int) and first is None and last is None:
            first = page_size

        if after is not None and (not isinstance(after, int) or after < 0):
            msg = "The node pointed with `after` does not exist."
            raise GraphQLPaginationArgumentValidationError(msg)

        if before is not None and (not isinstance(before, int) or before < 0):
            msg = "The node pointed with `before` does not exist."
            raise GraphQLPaginationArgumentValidationError(msg)

        if after is not None and before is not None and after >= before:
            msg = "The node pointed with `after` must be before the node pointed with `before`."
            raise GraphQLPaginationArgumentValidationError(msg)

        # Since `after` is also exclusive, we need to add 1 to it, so that slicing works correctly.
        if after is not None:
            after += 1

        return ValidatedPaginationArgs(after=after, before=before, first=first, last=last)

    def paginate_queryset(self, queryset: QuerySet, info: GQLInfo) -> QuerySet:
        """Paginate a top-level queryset."""
        self.calculate_pagination_arguments(queryset, info)
        return self.apply_pagination(queryset, info)

    def calculate_pagination_arguments(self, queryset: QuerySet, info: GQLInfo) -> None:
        """
        Calculates the pagination arguments for a top-level queryset.

        This function is based on the Relay pagination algorithm.
        See. https://relay.dev/graphql/connections.htm#sec-Pagination-algorithm
        """
        if self.requires_total_count:
            self.total_count = queryset.count()

        if self.after is not None:
            self.start = self.after

        if self.before is not None:
            self.stop = self.before

        if self.first is not None:
            self.stop = self.start + self.first if self.stop is None else min(self.start + self.first, self.stop)

        if self.last is not None:
            if self.stop is None:
                if self.total_count is None:
                    self.total_count = queryset.count()
                self.stop = self.total_count
            self.start = max(self.stop - self.last, self.start)

    def apply_pagination(self, queryset: QuerySet, info: GQLInfo) -> QuerySet:
        """Paginate a top-level queryset using queryset slicing."""
        return queryset[self.start : self.stop]

    def paginate_prefetch_queryset(self, queryset: QuerySet, field: ToManyField, info: GQLInfo) -> QuerySet:
        """Paginate a prefetch queryset."""
        self.calculate_prefetch_pagination_arguments(queryset, field, info)
        return self.apply_prefetch_pagination(queryset, field, info)

    def calculate_prefetch_pagination_arguments(self, queryset: QuerySet, field: ToManyField, info: GQLInfo) -> None:
        """
        Calculates the pagination arguments for a prefetch queryset.

        This function is based on the Relay pagination algorithm.
        See. https://relay.dev/graphql/connections.htm#sec-Pagination-algorithm
        """
        if self.requires_total_count:
            self.total_count = F(undine_settings.PAGINATION_TOTAL_COUNT_KEY)

        if self.after is not None:
            self.start = self.after

        if self.before is not None:
            self.stop = self.before

        if self.first is not None:
            self.stop = self.start + self.first if self.stop is None else min(self.start + self.first, self.stop)

        if self.last is not None:
            if self.stop is None:
                if self.total_count is None:
                    self.total_count = F(undine_settings.PAGINATION_TOTAL_COUNT_KEY)  # type: ignore[assignment]
                self.start = Greatest(self.total_count - Value(self.last), Value(self.start))  # type: ignore[assignment]
            else:
                self.start = max(self.stop - self.last, self.start)

    def apply_prefetch_pagination(self, queryset: QuerySet, field: ToManyField, info: GQLInfo) -> QuerySet:
        """
        Paginate a prefetch queryset using a window function partitioned by the given related field.

        Pagination arguments are annotated to the queryset, since they are calculated in the database.
        There is the issue that they might not be available if the queryset is empty after pagination,
        but since they can be different for each prefetch partition, we cannot do anything about that.
        """
        if isinstance(field, ManyToManyField | ManyToManyRel):
            register_for_prefetch_hack(queryset, field)

        related_name = field.remote_field.name

        if self.total_count is not None:
            queryset = _add_total_count(queryset, related_name)

        queryset = _add_partition_index(queryset, related_name)

        queryset = _add_start_index(queryset, self.start)
        queryset = _filter_by_start_index(queryset)

        if self.stop is not None:
            queryset = _add_stop_index(queryset, self.stop)
            queryset = _filter_by_stop_index(queryset)

        return queryset


class OffsetPagination:
    """A wrapper for paginating a `QueryType` using offset and limit."""

    def __init__(
        self,
        ref: type[QueryType | UnionType | InterfaceType],
        /,
        *,
        page_size: int | None = undine_settings.PAGINATION_PAGE_SIZE,
        pagination_handler: type[PaginationHandler] = PaginationHandler,
        description: str | None = None,
    ) -> None:
        """
        Create a new OffsetPagination.

        :param ref: The `QueryType`, `UnionType`, or `InterfaceType` to paginate.
        :param page_size: Maximum number of items to return in a page. No limit if `None`.
        :param pagination_handler: Handler to use for pagination.
        :param description: Description for the created GraphQL type.
        """
        self.query_type = ref if is_subclass(ref, QueryType) else None
        self.union_type = ref if is_subclass(ref, UnionType) else None
        self.interface_type = ref if is_subclass(ref, InterfaceType) else None

        self.page_size = page_size
        self.pagination_handler = pagination_handler
        self.description = description


def _add_partition_index(queryset: QuerySet, related_name: str) -> QuerySet:
    """Add an index to each instance in the queryset, partitioned by the given related name."""
    return queryset.alias(
        **{
            undine_settings.PAGINATION_INDEX_KEY: (
                Window(
                    expression=RowNumber(),
                    partition_by=F(related_name),
                    order_by=queryset.query.order_by or copy(queryset.model._meta.ordering) or None,
                )
                - Value(1)  # Start from zero.
            ),
        },
    )


def _add_total_count(queryset: QuerySet, related_name: str) -> QuerySet:
    """Add an annotation to the given queryset with the total count of objects in the queryset."""
    total_count = _total_count_subquery(queryset, related_name)
    return queryset.annotate(**{undine_settings.PAGINATION_TOTAL_COUNT_KEY: total_count})


def _total_count_subquery(queryset: QuerySet, related_name: str) -> SubqueryCount:
    """Get a subquery for calculating total count, partitioned by the given related name."""
    return SubqueryCount(queryset=queryset.filter(**{related_name: OuterRef(related_name)}))


def _add_start_index(queryset: QuerySet, start: int | CombinableExpression) -> QuerySet:
    """Add an annotation to the given queryset with the start index of the current page."""
    if isinstance(start, int):
        start = Value(start)
    return queryset.annotate(**{undine_settings.PAGINATION_START_INDEX_KEY: start})


def _filter_by_start_index(queryset: QuerySet) -> QuerySet:
    """Filter out all items before the start index of the current page."""
    start = F(undine_settings.PAGINATION_START_INDEX_KEY)
    return queryset.filter(**{f"{undine_settings.PAGINATION_INDEX_KEY}__gte": start})


def _add_stop_index(queryset: QuerySet, stop: int | CombinableExpression) -> QuerySet:
    """Add an annotation to the given queryset with the stop index of the current page."""
    if isinstance(stop, int):
        stop = Value(stop)
    return queryset.annotate(**{undine_settings.PAGINATION_STOP_INDEX_KEY: stop})


def _filter_by_stop_index(queryset: QuerySet) -> QuerySet:
    """Filter out all items on or after the stop index of the current page."""
    stop = F(undine_settings.PAGINATION_STOP_INDEX_KEY)
    return queryset.filter(**{f"{undine_settings.PAGINATION_INDEX_KEY}__lt": stop})



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphql import (
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    GraphQLError,
    InlineFragmentNode,
    OperationDefinitionNode,
    ValidationRule,
)

from undine.settings import undine_settings

if TYPE_CHECKING:
    from graphql import ExecutableDefinitionNode, SelectionNode, ValidationContext, VisitorAction


__all__ = [
    "MaxDirectiveCountRule",
]


class MaxDirectiveCountRule(ValidationRule):
    """Validates that the number of directives in a GraphQL query does not exceed the maximum allowed."""

    def __init__(self, context: ValidationContext) -> None:
        super().__init__(context)
        self.visited_fragments: set[str] = set()

    def enter_operation_definition(self, node: OperationDefinitionNode, *_args: Any) -> VisitorAction:
        directive_count = self.count_directives(node)

        if directive_count > undine_settings.MAX_ALLOWED_DIRECTIVES:
            msg = (
                f"Operation has more than {undine_settings.MAX_ALLOWED_DIRECTIVES} directives, "
                f"which exceeds the maximum allowed."
            )
            error = GraphQLError(msg, node)
            self.context.report_error(error)
            return self.BREAK

        return self.IDLE

    def count_directives(self, node: ExecutableDefinitionNode | SelectionNode) -> int:  # noqa: C901,PLR0912
        directive_count: int = 0

        match node:
            case FieldNode():
                directive_count += len(node.directives)

                if node.selection_set is not None:
                    for selection in node.selection_set.selections:
                        directive_count += self.count_directives(selection)

            case OperationDefinitionNode():
                for variable_definition in node.variable_definitions or ():
                    directive_count += len(variable_definition.directives)

                directive_count += len(node.directives)

                for selection in node.selection_set.selections:
                    directive_count += self.count_directives(selection)

            case InlineFragmentNode():
                directive_count += len(node.directives)

                for selection in node.selection_set.selections:
                    directive_count += self.count_directives(selection)

            case FragmentSpreadNode():
                directive_count += len(node.directives)

                if node.name.value not in self.visited_fragments:
                    self.visited_fragments.add(node.name.value)
                    fragment = self.context.get_fragment(node.name.value)
                    if fragment is not None:
                        directive_count += self.count_directives(fragment)

            case FragmentDefinitionNode():
                for variable_definition in node.variable_definitions or ():
                    directive_count += len(variable_definition.directives)

                directive_count += len(node.directives)

                for selection in node.selection_set.selections:
                    directive_count += self.count_directives(selection)

        return directive_count



from __future__ import annotations

import dataclasses
import re
import unicodedata
import urllib.parse
from collections import UserString
from typing import TYPE_CHECKING, Unpack

from django.contrib.postgres.search import SearchQuery, SearchVector
from django.db.models import Q
from django.utils.translation import get_language_from_path, get_language_from_request

from undine.settings import undine_settings

from .reflection import get_members

if TYPE_CHECKING:
    from undine import Filter
    from undine.typing import FTSLang, GQLInfo, LangCode, LangSep, PostgresFTSLangSpecificFields

__all__ = [
    "PostgresFTS",
]


class PostgresFTS:
    """
    Filter reference for Postgres full text search.

    Creates a SearchVector for the request language, annotates it to the queryset,
    and searches using a raw search string created using the `build_pg_search` function.
    """

    def __init__(
        self,
        *common: str,
        separator: LangSep = "|",
        **lang_specific: Unpack[PostgresFTSLangSpecificFields],
    ) -> None:
        """
        Postgres full text search filter reference.

        :param common: Common fields for all languages.
        :param separator: Separator for each search term in the query.
        :param lang_specific: Fields specific to each language.
        """
        self.separator = separator
        self.vectors: dict[FTSLang, SearchVector] = {
            lang: SearchVector(*set(*common, *fields), config=lang)  # type: ignore[misc]
            for lang, fields in lang_specific.items()
        }

    def get_search_language(self, info: GQLInfo) -> SearchLanguage:
        lang = get_request_search_language(info)
        if lang not in self.vectors:
            return TextSearchLang.ENGLISH
        return lang

    def get_vector_alias_key(self, ftr: Filter, lang: SearchLanguage) -> str:
        return f"{undine_settings.PG_TEXT_SEARCH_PREFIX}_{ftr.name}_{lang.name}"


@dataclasses.dataclass(frozen=True, slots=True)
class PostgresFTSExpressionResolver:
    """Resolves a filter using a Postgres full text search."""

    fts: PostgresFTS

    def __call__(self, root: Filter, info: GQLInfo, *, value: str) -> Q:
        lang = self.fts.get_search_language(info)
        key = self.fts.get_vector_alias_key(root, lang)
        search = build_pg_search(value, separator=self.fts.separator)
        query = {key: SearchQuery(value=search, config=lang.name, search_type="raw")}
        return Q(**query)


class SearchLanguage(UserString):
    """Postgres search language with the corresponding language code."""

    def __init__(self, name: FTSLang, *, code: LangCode) -> None:
        self.name = name
        self.code = code
        super().__init__(name)


class TextSearchLang:
    """Default postgres text search language configurations."""

    ARABIC = SearchLanguage("arabic", code="ar")
    ARMENIAN = SearchLanguage("armenian", code="hy")
    BASQUE = SearchLanguage("basque", code="eu")
    CATALAN = SearchLanguage("catalan", code="ca")
    DANISH = SearchLanguage("danish", code="da")
    DUTCH = SearchLanguage("dutch", code="nl")
    ENGLISH = SearchLanguage("english", code="en")
    FINNISH = SearchLanguage("finnish", code="fi")
    FRENCH = SearchLanguage("french", code="fr")
    GERMAN = SearchLanguage("german", code="de")
    GREEK = SearchLanguage("greek", code="el")
    HINDI = SearchLanguage("hindi", code="hi")
    HUNGARIAN = SearchLanguage("hungarian", code="hu")
    INDONESIAN = SearchLanguage("indonesian", code="id")
    IRISH = SearchLanguage("irish", code="ga")
    ITALIAN = SearchLanguage("italian", code="it")
    LITHUANIAN = SearchLanguage("lithuanian", code="lt")
    NEPALI = SearchLanguage("nepali", code="ne")
    NORWEGIAN = SearchLanguage("norwegian", code="nb")
    PORTUGUESE = SearchLanguage("portuguese", code="pt")
    ROMANIAN = SearchLanguage("romanian", code="ro")
    RUSSIAN = SearchLanguage("russian", code="ru")
    SERBIAN = SearchLanguage("serbian", code="sr")
    SPANISH = SearchLanguage("spanish", code="es")
    SWEDISH = SearchLanguage("swedish", code="sv")
    TAMIL = SearchLanguage("tamil", code="ta")
    TURKISH = SearchLanguage("turkish", code="tr")
    YIDDISH = SearchLanguage("yiddish", code="yi")

    @classmethod
    def members(cls) -> dict[str, SearchLanguage]:
        return get_members(cls, SearchLanguage)

    @classmethod
    def for_code(cls, code: LangCode, *, default: SearchLanguage = ENGLISH) -> SearchLanguage:
        members = cls.members()
        for search_lang in members.values():
            if search_lang.code == code:
                return search_lang
        return default


def normalize_search_text(text: str) -> str:
    """
    Normalize the given text to use in full text search.

    Replaces all non-word and non-space characters with spaces and removes multiple spaces.
    Each string of chars separated by space then becomes a search term.
    Each search term is then quoted and matched as a possible start of word.
    """
    text = unicodedata.normalize("NFKC", text)  # Normalize text to unicode form
    text = re.sub(r"[^\w\s]", " ", text)  # Remove all non-word and non-space characters
    return re.sub(r"\s+", " ", text)  # Collapse multiple spaces


def build_pg_search(text: str, *, separator: LangSep = "|") -> str:
    """
    Build raw postgres full text search query from the given text.
    Text is normalized first by removing punctuation, etc.

    Each search term is matched together with other search terms using the given operator.
      | = Only one of the search terms needs to match
      & = All search terms need to match
      <-> = Each search term must be followed by the others
      <3> = Each search term must be followed by less than 3 "words" away (<1> same as <->)

    Ref. https://www.postgresql.org/docs/current/datatype-textsearch.html#DATATYPE-TSQUERY
    """
    norm_text = normalize_search_text(text)
    search_terms = (f"'{value}':*" for value in norm_text.split(" ") if value)
    return f" {separator} ".join(search_terms)


def get_request_search_language(info: GQLInfo) -> SearchLanguage:
    """
    Get language from the given request using this order:

    1. Referer header path
    2. Request info path
    3. Cookie as set by the LANGUAGE_COOKIE_NAME setting
    4. Accept-Language header
    5. Default language as set by the LANGUAGE_CODE setting
    """
    referer = info.context.META.get("HTTP_REFERER")
    if referer:
        path = urllib.parse.urlparse(referer).path
        lang: LangCode = get_language_from_path(path)  # type: ignore[assignment]
        if lang is not None:
            return TextSearchLang.for_code(lang)

    lang = get_language_from_request(info.context, check_path=True)  # type: ignore[assignment,arg-type]
    return TextSearchLang.for_code(lang)



from __future__ import annotations

import asyncio
import dataclasses
import io
import json
from collections.abc import AsyncIterator
from contextlib import suppress
from functools import cached_property, wraps
from inspect import isawaitable
from typing import TYPE_CHECKING, Any

from asgiref.typing import WebSocketAcceptEvent, WebSocketCloseEvent, WebSocketSendEvent
from django.core.handlers.asgi import ASGIRequest
from graphql import GraphQLError

from undine.exceptions import (
    GraphQLErrorGroup,
    GraphQLUnexpectedError,
    WebSocketConnectionInitAlreadyInProgressError,
    WebSocketConnectionInitForbiddenError,
    WebSocketConnectionInitTimeoutError,
    WebSocketEmptyMessageError,
    WebSocketError,
    WebSocketInternalServerError,
    WebSocketInvalidCompleteMessageOperationIdError,
    WebSocketInvalidConnectionInitPayloadError,
    WebSocketInvalidJSONError,
    WebSocketInvalidPingPayloadError,
    WebSocketInvalidPongPayloadError,
    WebSocketInvalidSubscribeOperationIdError,
    WebSocketInvalidSubscribePayloadError,
    WebSocketMissingCompleteMessageOperationIdError,
    WebSocketMissingSubscribeOperationIdError,
    WebSocketMissingSubscribePayloadError,
    WebSocketSubscriberForOperationIdAlreadyExistsError,
    WebSocketTooManyInitialisationRequestsError,
    WebSocketTypeMissingError,
    WebSocketUnauthorizedError,
    WebSocketUnknownMessageTypeError,
    WebSocketUnsupportedSubProtocolError,
)
from undine.execution import execute_graphql_websocket
from undine.parsers import GraphQLRequestParamsParser
from undine.settings import undine_settings
from undine.typing import CompleteMessage, ConnectionAckMessage, ErrorMessage, NextMessage, PongMessage

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from django.contrib.auth.models import AnonymousUser, User
    from django.contrib.sessions.backends.base import SessionBase
    from django.core.files.uploadedfile import UploadedFile
    from django.http import HttpHeaders, QueryDict
    from django.http.request import MediaType
    from django.utils.datastructures import MultiValueDict
    from graphql import ExecutionResult

    from undine.dataclasses import GraphQLHttpParams
    from undine.typing import (
        ClientMessage,
        ConnectionInitMessage,
        DjangoRequestProtocol,
        GraphQLWebSocketCloseCode,
        P,
        PingMessage,
        RequestMethod,
        ServerMessage,
        SubscribeMessage,
        WebSocketASGIScope,
        WebSocketProtocol,
    )


__all__ = [
    "GRAPHQL_TRANSPORT_WS_PROTOCOL",
    "GraphQLOverWebSocketHandler",
    "WebSocketOperation",
    "WebSocketRequest",
]


GRAPHQL_TRANSPORT_WS_PROTOCOL = "graphql-transport-ws"


def close_websocket_on_error(func: Callable[P, Awaitable[None]]) -> Callable[P, Awaitable[None]]:
    """Close websockets if exceptions are raised."""

    @wraps(func)
    async def wrapper(self: GraphQLOverWebSocketHandler, *args: P.args, **kwargs: P.kwargs) -> None:
        try:
            await func(self, *args, **kwargs)  # type: ignore[arg-type]

        except WebSocketError as error:
            await self.disconnect()
            await self.close(code=error.code, reason=error.reason)

        except Exception:  # noqa: BLE001  # pragma: no cover
            err = WebSocketInternalServerError
            await self.disconnect()
            await self.close(code=err.code, reason=err.reason)

    return wrapper  # type: ignore[return-value]


@dataclasses.dataclass(kw_only=True, slots=True)
class GraphQLOverWebSocketHandler:
    """
    Handler for the GraphQL over WebSocket Protocol.
    See: https://github.com/graphql/graphql-over-http/blob/main/rfcs/GraphQLOverWebSocket.md
    """

    websocket: WebSocketProtocol

    connection_init_timeout_task: asyncio.Task | None = None
    connection_init_timed_out: bool = False
    connection_init_received: bool = False
    connection_acknowledged: bool = False

    operations: dict[str, WebSocketOperation] = dataclasses.field(default_factory=dict)

    # Core methods

    @close_websocket_on_error
    async def connect(self) -> None:
        """Websocket created, start waiting for a connection init message."""
        if GRAPHQL_TRANSPORT_WS_PROTOCOL not in self.websocket.scope["subprotocols"]:
            raise WebSocketUnsupportedSubProtocolError

        # Shouldn't happen, but just in case.
        if isinstance(self.connection_init_timeout_task, asyncio.Task):
            self.connection_init_timeout_task.cancel()
            raise WebSocketConnectionInitAlreadyInProgressError

        self.connection_init_timeout_task = asyncio.create_task(self.handle_connection_init_timeout())
        await self.accept()

    @close_websocket_on_error
    async def receive(self, data: str | None = None) -> None:
        """Process a message received from the websocket."""
        message = self.validate_message(data)

        match message["type"]:
            case "connection_init":
                self.validate_connection_init_message(message)
                await self.handle_connection_init(message)

            case "ping":
                self.validate_ping_message(message)
                await self.handle_ping(message)

            case "pong":
                self.validate_pong_message(message)
                await self.handle_pong(message)

            case "subscribe":
                self.validate_subscribe_message(message)
                await self.handle_subscribe(message)

            case "complete":
                self.validate_complete_message(message)
                await self.handle_complete(message)

            case _:
                raise WebSocketUnknownMessageTypeError(type=message["type"])

    async def disconnect(self) -> None:
        """Await all done tasks and cancel all pending tasks before closing the websocket."""
        if self.connection_init_timeout_task is not None:
            if not self.connection_init_timeout_task.done():
                self.connection_init_timeout_task.cancel()

            with suppress(BaseException):
                await self.connection_init_timeout_task

        for operation_id in list(self.operations):
            operation = self.operations.pop(operation_id)
            if not operation.task.done():
                operation.task.cancel()

            with suppress(BaseException):
                await operation.task

    async def accept(self) -> None:
        event = WebSocketAcceptEvent(type="websocket.accept", subprotocol=GRAPHQL_TRANSPORT_WS_PROTOCOL, headers=[])
        await self.websocket.send(message=event)

    async def close(self, code: GraphQLWebSocketCloseCode, reason: str) -> None:
        event = WebSocketCloseEvent(type="websocket.close", code=code, reason=reason)
        await self.websocket.send(message=event)

    async def send(self, message: ServerMessage) -> None:
        text = json.dumps(message, separators=(",", ":"))
        event = WebSocketSendEvent(type="websocket.send", text=text, bytes=None)
        await self.websocket.send(message=event)

    # Message validation

    def validate_message(self, text_data: str | None) -> ClientMessage:
        if text_data is None:
            raise WebSocketEmptyMessageError

        try:
            message = json.loads(text_data)
        except Exception as error:
            raise WebSocketInvalidJSONError from error

        if not isinstance(message, dict):
            raise WebSocketInvalidJSONError

        if "type" not in message:
            raise WebSocketTypeMissingError

        return message  # type: ignore[return-value]

    def validate_connection_init_message(self, message: ConnectionInitMessage) -> None:
        if "payload" in message and not isinstance(message["payload"], dict):
            raise WebSocketInvalidConnectionInitPayloadError

    def validate_ping_message(self, message: PingMessage) -> None:
        if "payload" in message and not isinstance(message["payload"], dict):
            raise WebSocketInvalidPingPayloadError

    def validate_pong_message(self, message: PongMessage) -> None:
        if "payload" in message and not isinstance(message["payload"], dict):
            raise WebSocketInvalidPongPayloadError

    def validate_subscribe_message(self, message: SubscribeMessage) -> None:
        if "id" not in message:
            raise WebSocketMissingSubscribeOperationIdError

        if not isinstance(message["id"], str):
            raise WebSocketInvalidSubscribeOperationIdError

        if "payload" not in message:
            raise WebSocketMissingSubscribePayloadError

        if not isinstance(message["payload"], dict):
            raise WebSocketInvalidSubscribePayloadError

    def validate_complete_message(self, message: CompleteMessage) -> None:
        if "id" not in message:
            raise WebSocketMissingCompleteMessageOperationIdError

        if not isinstance(message["id"], str):
            raise WebSocketInvalidCompleteMessageOperationIdError

    # Message handlers

    async def handle_connection_init(self, message: ConnectionInitMessage) -> None:
        """Indicates that the client wants to establish a connection within the existing socket."""
        if self.connection_init_timed_out:
            return

        if self.connection_init_timeout_task is not None:
            self.connection_init_timeout_task.cancel()

        if self.connection_init_received:
            raise WebSocketTooManyInitialisationRequestsError

        self.connection_init_received = True

        request = WebSocketRequest(scope=self.websocket.scope, message=message)
        try:
            payload = undine_settings.WEBSOCKET_CONNECTION_INIT_HOOK(request)
            if isawaitable(payload):
                payload = await payload
        except GraphQLError as error:
            raise WebSocketConnectionInitForbiddenError(reason=error.message) from error
        except Exception as error:
            raise WebSocketConnectionInitForbiddenError from error

        response = ConnectionAckMessage(type="connection_ack")
        if payload is not None:
            response["payload"] = payload

        await self.send(message=response)

        self.connection_acknowledged = True

    async def handle_connection_init_timeout(self) -> None:
        """Handle connection init timeout. Started as a task when a socket is accepted."""
        await asyncio.sleep(delay=undine_settings.WEBSOCKET_CONNECTION_INIT_TIMEOUT_SECONDS)
        if self.connection_init_received:
            return

        self.connection_init_timed_out = True

        error = WebSocketConnectionInitTimeoutError
        await self.close(code=error.code, reason=error.reason)

    async def handle_ping(self, message: PingMessage) -> None:
        """
        Useful for detecting failed connections, displaying latency metrics or other types of network probing.

        A Pong must be sent in response from the receiving party as soon as possible.
        The Ping message can be sent at any time within the established socket.
        """
        request = WebSocketRequest(scope=self.websocket.scope, message=message)
        payload = undine_settings.WEBSOCKET_PING_HOOK(request)
        if isawaitable(payload):
            payload = await payload

        response = PongMessage(type="pong")
        if payload is not None:
            response["payload"] = payload

        await self.send(message=response)

    async def handle_pong(self, message: PongMessage) -> None:
        """
        The Pong message can be sent at any time within the established socket,
        and may even be sent unsolicited as an unidirectional heartbeat.
        """
        request = WebSocketRequest(scope=self.websocket.scope, message=message)
        hook = undine_settings.WEBSOCKET_PONG_HOOK(request)
        if isawaitable(hook):
            await hook

    async def handle_subscribe(self, message: SubscribeMessage) -> None:
        """
        Request an operation specified in the message payload. This message provides a unique ID
        field to connect published messages to the operation requested by this message.
        """
        if not self.connection_acknowledged:
            raise WebSocketUnauthorizedError

        if message["id"] in self.operations:
            raise WebSocketSubscriberForOperationIdAlreadyExistsError(id=message["id"])

        try:
            params = GraphQLRequestParamsParser.get_graphql_params(message["payload"])
        except GraphQLError as error:
            raise WebSocketInvalidSubscribePayloadError(reason=error.message) from error

        self.operations[message["id"]] = WebSocketOperation(
            handler=self,
            operation_id=message["id"],
            params=params,
            request=WebSocketRequest(scope=self.websocket.scope, message=message),  # type: ignore[arg-type]
        )

    async def handle_complete(self, message: CompleteMessage) -> None:
        """
        Client has stopped listening and wants to complete the subscription.
        No further events, relevant to the original subscription, should be sent through.
        """
        operation = self.operations.get(message["id"])
        if operation is not None:
            operation.set_completed()
            operation.task.cancel()

            with suppress(BaseException):
                await operation.task


@dataclasses.dataclass(kw_only=True, slots=True)
class WebSocketOperation:
    """Encapsulates a GraphQL operation received through a WebSocket."""

    handler: GraphQLOverWebSocketHandler
    operation_id: str
    params: GraphQLHttpParams
    request: DjangoRequestProtocol

    task: asyncio.Task = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        # Operation starts executing as soon an instance is created.
        self.task = asyncio.create_task(self.run())

    @property
    def is_completed(self) -> bool:
        return self.operation_id not in self.handler.operations

    def set_completed(self) -> None:
        self.handler.operations.pop(self.operation_id, None)

    async def run(self) -> None:
        result = await execute_graphql_websocket(self.params, self.request)

        if isinstance(result, AsyncIterator):
            await self.execute_subscription(result)
            return

        await self.execute_singe_result_operation(result)

    async def execute_singe_result_operation(self, result: ExecutionResult) -> None:
        if result.errors:
            await self.send_errors(errors=result.errors)
            return

        await self.send_next(result=result)
        await self.send_complete()

    async def execute_subscription(self, source: AsyncIterator[ExecutionResult]) -> None:
        initial = True
        try:
            async for result in source:
                if initial and result.errors:
                    await self.send_errors(errors=result.errors)
                    return

                initial = False
                await self.send_next(result=result)

        except GraphQLError as error:
            await self.send_errors(errors=[error])

        except GraphQLErrorGroup as error:
            await self.send_errors(errors=list(error.flatten()))

        except Exception as error:  # noqa: BLE001
            await self.send_errors(errors=[GraphQLUnexpectedError(message=str(error))])

        else:
            await self.send_complete()

    async def send_errors(self, errors: list[GraphQLError]) -> None:
        """
        Operation execution error(s) in response to the Subscribe message.
        This can occur before execution starts, usually due to validation errors,
        or during the execution of the request. This message terminates the operation
        and no further messages will be sent.
        """
        if self.is_completed:  # pragma: no cover
            return

        error_payload = [err.formatted for err in errors]
        error_message = ErrorMessage(type="error", id=self.operation_id, payload=error_payload)
        await self.handler.send(error_message)

        self.set_completed()

    async def send_next(self, result: ExecutionResult) -> None:
        """
        Operation execution result(s) from the source stream created by the binding Subscribe message.
        After all results have been emitted, the Complete message will follow indicating stream completion.
        """
        if self.is_completed:  # pragma: no cover
            return

        next_message = NextMessage(type="next", id=self.operation_id, payload=result.formatted)
        await self.handler.send(next_message)

    async def send_complete(self) -> None:
        """The requested operation execution has completed."""
        if self.is_completed:  # pragma: no cover
            return

        complete_message = CompleteMessage(type="complete", id=self.operation_id)
        await self.handler.send(complete_message)

        self.set_completed()


@dataclasses.dataclass(kw_only=True)  # No slots due to '@cached_property'
class WebSocketRequest:
    """Imitate a Django HttpRequest object from a WebSocket connection."""

    scope: WebSocketASGIScope
    message: ConnectionInitMessage | PingMessage | PongMessage | SubscribeMessage

    @cached_property
    def _request(self) -> ASGIRequest:
        body = json.dumps(self.message["payload"]).encode("utf-8")
        # Method is not technically part of WebSocketASGIScope,
        # but needs to be set for ASGIRequest to initialize correctly.
        self.scope["method"] = "WEBSOCKET"  # type: ignore[typeddict-unknown-key]
        return ASGIRequest(scope=self.scope, body_file=io.BytesIO(body))

    @property
    def GET(self) -> QueryDict:  # noqa: N802
        return self._request.GET

    @property
    def POST(self) -> QueryDict:  # noqa: N802
        return self._request.POST

    @property
    def COOKIES(self) -> dict[str, str]:  # noqa: N802
        return self._request.COOKIES

    @property
    def FILES(self) -> MultiValueDict[str, UploadedFile]:  # noqa: N802
        return self._request.FILES

    @property
    def META(self) -> dict[str, Any]:  # noqa: N802
        return self._request.META

    @property
    def scheme(self) -> str | None:
        return self._request.scheme

    @property
    def path(self) -> str:
        return self._request.path

    @property
    def method(self) -> RequestMethod:
        return self._request.method  # type: ignore[return-value]

    @property
    def headers(self) -> HttpHeaders:
        return self._request.headers

    @property
    def body(self) -> bytes:
        return self._request.body

    @property
    def encoding(self) -> str | None:
        return self._request.encoding

    @property
    def user(self) -> User | AnonymousUser:
        return self.scope["user"]

    async def auser(self) -> User | AnonymousUser:
        return self.user

    @property
    def session(self) -> SessionBase:
        return self.scope["session"]

    @property
    def content_type(self) -> str | None:
        return self._request.content_type

    @property
    def content_params(self) -> dict[str, str] | None:
        return self._request.content_params

    @property
    def accepted_types(self) -> list[MediaType]:
        return self._request.accepted_types

    @property
    def response_content_type(self) -> str:
        return "application/json"

    @response_content_type.setter
    def response_content_type(self, value: str) -> None: ...


# Default hooks


def connection_init_hook(request: WebSocketRequest) -> dict[str, Any] | None:
    """Default hook for custom connection init handling."""


def ping_hook(request: WebSocketRequest) -> dict[str, Any] | None:
    """Default hook for custom ping handling."""


def pong_hook(request: WebSocketRequest) -> None:
    """Default hook for custom pong handling."""



from __future__ import annotations

import datetime
import decimal
import uuid
from enum import Enum
from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel, GenericRelation
from django.db.models import F, Model, TextChoices
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute

from undine import Input, MutationType
from undine.converters import convert_to_input_ref
from undine.dataclasses import LazyLambda, TypeRef
from undine.exceptions import InvalidInputMutationTypeError, ModelFieldDoesNotExistError
from undine.typing import GenericField, Lambda, ModelField, MutationKind, RelatedField, RelationType
from undine.utils.model_utils import generic_foreign_key_for_generic_relation, get_model_field, get_related_name


@convert_to_input_ref.register
def _(_: None, **kwargs: Any) -> Any:
    caller: Input = kwargs["caller"]
    field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    return convert_to_input_ref(field, **kwargs)


@convert_to_input_ref.register
def _(ref: F, **kwargs: Any) -> Any:
    caller: Input = kwargs["caller"]
    field = get_model_field(model=caller.mutation_type.__model__, lookup=ref.name)
    return convert_to_input_ref(field, **kwargs)


@convert_to_input_ref.register
def _(ref: ModelField, **kwargs: Any) -> Any:
    return ref


@convert_to_input_ref.register
def _(ref: RelatedField, **kwargs: Any) -> Any:
    return ref.related_model


@convert_to_input_ref.register
def _(ref: type[Model], **kwargs: Any) -> Any:
    return ref


@convert_to_input_ref.register
def _(ref: DeferredAttribute | ForwardManyToOneDescriptor, **kwargs: Any) -> Any:
    return ref.field


@convert_to_input_ref.register
def _(ref: ReverseManyToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_input_ref(ref.rel, **kwargs)


@convert_to_input_ref.register
def _(ref: ReverseOneToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_input_ref(ref.related, **kwargs)


@convert_to_input_ref.register
def _(ref: ManyToManyDescriptor, **kwargs: Any) -> Any:
    return convert_to_input_ref(ref.rel if ref.reverse else ref.field, **kwargs)


@convert_to_input_ref.register
def _(ref: str, **kwargs: Any) -> Any:
    caller: Input = kwargs["caller"]
    field = get_model_field(model=caller.mutation_type.__model__, lookup=ref)
    return convert_to_input_ref(field, **kwargs)


@convert_to_input_ref.register
def _(ref: Lambda, **kwargs: Any) -> Any:
    return LazyLambda(callback=ref)


@convert_to_input_ref.register
def _(ref: type[str], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[bool], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[int], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[float], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[decimal.Decimal], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[datetime.datetime], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[datetime.date], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[datetime.time], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[datetime.timedelta], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[uuid.UUID], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[Enum], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[TextChoices], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[list], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: type[dict], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_input_ref.register
def _(ref: FunctionType, **kwargs: Any) -> Any:
    return ref


@convert_to_input_ref.register
def _(ref: type[MutationType], **kwargs: Any) -> Any:
    if ref.__kind__ != MutationKind.related:
        raise InvalidInputMutationTypeError(ref=ref, kind=ref.__kind__)

    caller: Input = kwargs["caller"]

    model = caller.mutation_type.__model__

    try:
        field: RelatedField | GenericField = get_model_field(model=model, lookup=caller.field_name)  # type: ignore[assignment]
    except ModelFieldDoesNotExistError:
        # Allow using relations "wrong", e.g. provide single dict for many related field.
        # Actual mutation for related mutations will be handled by the user.
        return ref

    relation_type = RelationType.for_related_field(field)

    # Remove the Input for the reverse relation from the MutationType used as the Input for this one.
    if relation_type.is_generic_relation:
        reverse_field_name: str = generic_foreign_key_for_generic_relation(field).name  # type: ignore[arg-type]
        ref.__input_map__.pop(reverse_field_name, None)

    elif not relation_type.is_generic_foreign_key:
        reverse_field_name = get_related_name(field.remote_field)  # type: ignore[arg-type]
        ref.__input_map__.pop(reverse_field_name, None)

    return ref


@convert_to_input_ref.register
def _(ref: GenericRelation, **kwargs: Any) -> Any:
    return ref.related_model


@convert_to_input_ref.register
def _(ref: GenericRel, **kwargs: Any) -> Any:
    return convert_to_input_ref(ref.field, **kwargs)


@convert_to_input_ref.register
def _(ref: GenericForeignKey, **kwargs: Any) -> Any:
    return ref



from __future__ import annotations

import inspect
import json
from collections.abc import Awaitable, Callable
from functools import wraps
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, TypeAlias, overload

from django.http import HttpRequest, HttpResponse
from django.http.request import MediaType

from undine.exceptions import (
    GraphQLMissingContentTypeError,
    GraphQLRequestDecodingError,
    GraphQLUnsupportedContentTypeError,
)
from undine.integrations.graphiql import render_graphiql
from undine.settings import undine_settings
from undine.typing import DjangoRequestProtocol, DjangoResponseProtocol
from undine.utils.graphql.utils import get_error_execution_result

if TYPE_CHECKING:
    from collections.abc import Iterable

    from graphql import ExecutionResult

    from undine.typing import RequestMethod

__all__ = [
    "HttpMethodNotAllowedResponse",
    "HttpUnsupportedContentTypeResponse",
    "decode_body",
    "get_preferred_response_content_type",
    "graphql_result_response",
    "load_json_dict",
    "parse_json_body",
    "require_graphql_request",
    "require_persisted_documents_request",
]


class HttpMethodNotAllowedResponse(HttpResponse):
    def __init__(self, allowed_methods: Iterable[RequestMethod]) -> None:
        msg = "Method not allowed"
        super().__init__(content=msg, status=HTTPStatus.METHOD_NOT_ALLOWED, content_type="text/plain; charset=utf-8")
        self["Allow"] = ", ".join(allowed_methods)


class HttpUnsupportedContentTypeResponse(HttpResponse):
    def __init__(self, supported_types: Iterable[str]) -> None:
        msg = "Server does not support any of the requested content types."
        super().__init__(content=msg, status=HTTPStatus.NOT_ACCEPTABLE, content_type="text/plain; charset=utf-8")
        self["Accept"] = ", ".join(supported_types)


def get_preferred_response_content_type(accepted: list[MediaType], supported: list[str]) -> str | None:
    """Get the first supported media type matching given accepted types."""
    for accepted_type in accepted:
        for supported_type in supported:
            if accepted_type.match(supported_type):
                return supported_type
    return None


def parse_json_body(body: bytes, charset: str = "utf-8") -> dict[str, Any]:
    """
    Parse JSON body.

    :param body: The body to parse.
    :param charset: The charset to decode the body with.
    :raises GraphQLDecodeError: If the body cannot be decoded.
    :return: The parsed JSON body.
    """
    decoded = decode_body(body, charset=charset)
    return load_json_dict(
        decoded,
        decode_error_msg="Could not load JSON body.",
        type_error_msg="JSON body should convert to a dictionary.",
    )


def decode_body(body: bytes, charset: str = "utf-8") -> str:
    """
    Decode body.

    :param body: The body to decode.
    :param charset: The charset to decode the body with.
    :raises GraphQLRequestDecodingError: If the body cannot be decoded.
    :return: The decoded body.
    """
    try:
        return body.decode(encoding=charset)
    except Exception as error:
        msg = f"Could not decode body with encoding '{charset}'."
        raise GraphQLRequestDecodingError(msg) from error


def load_json_dict(string: str, *, decode_error_msg: str, type_error_msg: str) -> dict[str, Any]:
    """
    Load JSON dict from string, raising GraphQL errors if decoding fails.

    :param string: The string to load.
    :param decode_error_msg: The error message to use if decoding fails.
    :param type_error_msg: The error message to use if the string is not a JSON object.
    :raises GraphQLRequestDecodingError: If decoding fails or the string is not a JSON object.
    :return: The loaded JSON dict.
    """
    try:
        data = json.loads(string)
    except Exception as error:
        raise GraphQLRequestDecodingError(decode_error_msg) from error

    if not isinstance(data, dict):
        raise GraphQLRequestDecodingError(type_error_msg)
    return data


def graphql_result_response(
    result: ExecutionResult,
    *,
    status: int = HTTPStatus.OK,
    content_type: str = "application/json",
) -> DjangoResponseProtocol:
    """Serialize the given execution result to an HTTP response."""
    content = json.dumps(result.formatted, separators=(",", ":"))
    return HttpResponse(content=content, status=status, content_type=content_type)  # type: ignore[return-value]


SyncViewIn: TypeAlias = Callable[[DjangoRequestProtocol], DjangoResponseProtocol]
AsyncViewIn: TypeAlias = Callable[[DjangoRequestProtocol], Awaitable[DjangoResponseProtocol]]

SyncViewOut: TypeAlias = Callable[[HttpRequest], HttpResponse]
AsyncViewOut: TypeAlias = Callable[[HttpRequest], Awaitable[HttpResponse]]


@overload
def require_graphql_request(func: SyncViewIn) -> SyncViewOut: ...


@overload
def require_graphql_request(func: AsyncViewIn) -> AsyncViewOut: ...


def require_graphql_request(func: SyncViewIn | AsyncViewIn) -> SyncViewOut | AsyncViewOut:
    """
    Perform various checks on the request to ensure it's suitable for GraphQL operations.
    Can also return early to display GraphiQL.
    """
    methods: list[RequestMethod] = ["GET", "POST"]

    def get_supported_types() -> list[str]:
        supported_types = ["application/graphql-response+json", "application/json"]
        if undine_settings.GRAPHIQL_ENABLED:
            supported_types.append("text/html")
        return supported_types

    if inspect.iscoroutinefunction(func):

        @wraps(func)
        async def wrapper(request: DjangoRequestProtocol) -> DjangoResponseProtocol | HttpResponse:
            if request.method not in methods:
                return HttpMethodNotAllowedResponse(allowed_methods=methods)

            supported_types = get_supported_types()
            media_type = get_preferred_response_content_type(accepted=request.accepted_types, supported=supported_types)
            if media_type is None:
                return HttpUnsupportedContentTypeResponse(supported_types=supported_types)

            if media_type == "text/html":
                return render_graphiql(request)  # type: ignore[arg-type]

            request.response_content_type = media_type
            return await func(request)

    else:

        @wraps(func)
        def wrapper(request: DjangoRequestProtocol) -> DjangoResponseProtocol | HttpResponse:
            if request.method not in methods:
                return HttpMethodNotAllowedResponse(allowed_methods=methods)

            supported_types = get_supported_types()
            media_type = get_preferred_response_content_type(accepted=request.accepted_types, supported=supported_types)
            if media_type is None:
                return HttpUnsupportedContentTypeResponse(supported_types=supported_types)

            if media_type == "text/html":
                return render_graphiql(request)  # type: ignore[arg-type]

            request.response_content_type = media_type
            return func(request)  # type: ignore[return-value]

    return wrapper  # type: ignore[return-value]


def require_persisted_documents_request(func: SyncViewIn) -> SyncViewOut:
    """Perform various checks on the request to ensure that it's suitable for registering persisted documents."""
    content_type: str = "application/json"
    methods: list[RequestMethod] = ["POST"]

    @wraps(func)
    def wrapper(request: DjangoRequestProtocol) -> DjangoResponseProtocol | HttpResponse:
        if request.method not in methods:
            return HttpMethodNotAllowedResponse(allowed_methods=methods)

        media_type = get_preferred_response_content_type(accepted=request.accepted_types, supported=[content_type])
        if media_type is None:
            return HttpUnsupportedContentTypeResponse(supported_types=[content_type])

        request.response_content_type = media_type

        if request.content_type is None:  # pragma: no cover
            result = get_error_execution_result(GraphQLMissingContentTypeError())
            return graphql_result_response(result, status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE, content_type=media_type)

        if not MediaType(request.content_type).match(content_type):
            result = get_error_execution_result(GraphQLUnsupportedContentTypeError(content_type=request.content_type))
            return graphql_result_response(result, status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE, content_type=media_type)

        return func(request)

    return wrapper  # type: ignore[return-value]



from __future__ import annotations

from abc import ABC, abstractmethod
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Self, Unpack

from graphql import DirectiveLocation, GraphQLArgument, Undefined

from undine.converters import convert_to_graphql_type
from undine.dataclasses import TypeRef
from undine.exceptions import (
    GraphQLMissingCalculationArgumentError,
    GraphQLUnexpectedCalculationArgumentError,
    MissingCalculationReturnTypeError,
)
from undine.parsers import parse_class_attribute_docstrings
from undine.settings import undine_settings
from undine.typing import T_co
from undine.utils.graphql.utils import check_directives
from undine.utils.reflection import get_members, get_wrapped_func
from undine.utils.text import dotpath, to_schema_name

if TYPE_CHECKING:
    from graphql import GraphQLInputType

    from undine.directives import Directive
    from undine.typing import (
        CalculationArgumentParams,
        DefaultValueType,
        DjangoExpression,
        GQLInfo,
        TypeHint,
        VisibilityFunc,
    )

__all__ = [
    "Calculation",
    "CalculationArgument",
]


class Calculation(ABC, Generic[T_co]):
    """
    An object that wraps logic for calculating a field's value based on defined input arguments.

    >>> class ExampleCalculation(Calculation[int]):
    ...     value = CalculationArgument(int)
    ...
    ...     def __call__(self, info: GQLInfo) -> DjangoExpression:
    ...         return Value(self.value)
    >>>
    >>> class TaskType(QueryType[Task]): ...
    >>>     example = Field(ExampleCalculation)
    """

    # Members should use `__dunder__` names to avoid name collisions with possible `CalculationArgument` names.

    # Set in '__init_subclass__'
    __returns__: ClassVar[TypeHint]
    __arguments__: ClassVar[dict[str, CalculationArgument]]
    __attribute_docstrings__: ClassVar[dict[str, str]]

    @abstractmethod
    def __call__(self, info: GQLInfo) -> DjangoExpression:
        """Calculate the value of the Field. Return an expression that can be annotated to a queryset."""

    def __class_getitem__(cls, returns: TypeHint) -> type[Calculation[TypeHint]]:
        cls.__returns__ = returns
        return cls  # type: ignore[return-value]

    def __init_subclass__(cls) -> None:
        try:
            cls.__returns__ = Calculation.__returns__
            del Calculation.__returns__
        except AttributeError as error:
            raise MissingCalculationReturnTypeError(name=cls.__name__) from error

        cls.__arguments__ = get_members(cls, CalculationArgument)
        cls.__attribute_docstrings__ = parse_class_attribute_docstrings(cls)

        for name, arg in cls.__arguments__.items():
            arg.__connect__(cls, name)

    def __init__(self, __field_name__: str, /, **kwargs: Any) -> None:
        parameters: dict[str, Any] = {}

        for arg in self.__arguments__.values():
            value = kwargs.pop(arg.name, arg.default_value)
            if value is Undefined:
                raise GraphQLMissingCalculationArgumentError(arg=arg.name, name=__field_name__)

            parameters[arg.name] = value

        if kwargs:
            raise GraphQLUnexpectedCalculationArgumentError(name=__field_name__, kwargs=kwargs)

        self.__field_name__: str = __field_name__
        self.__parameters__: MappingProxyType[str, Any] = MappingProxyType(parameters)


class CalculationArgument:
    """Defines an input argument for a `Calculation`."""

    def __init__(self, ref: TypeHint, **kwargs: Unpack[CalculationArgumentParams]) -> None:
        """
        Create a new `CalculationArgument`.

        :param ref: The argument reference to use for the `CalculationArgument`.
        :param default_value: The default value for the `CalculationArgument`.
        :param description: Description for the `CalculationArgument`.
        :param deprecation_reason: If the `CalculationArgument` is deprecated, describes the reason for deprecation.
        :param schema_name: Actual name in the GraphQL schema. Only needed if argument name is a python keyword.
        :param directives: GraphQL directives for the `CalculationArgument`.
        :param extensions: GraphQL extensions for the `CalculationArgument`.
        """
        self.ref = ref

        self.default_value: DefaultValueType = kwargs.get("default_value", Undefined)
        self.description: str | None = kwargs.get("description", Undefined)  # type: ignore[assignment]
        self.deprecation_reason: str | None = kwargs.get("deprecation_reason")  # type: ignore[assignment]
        self.schema_name: str = kwargs.get("schema_name", Undefined)  # type: ignore[assignment]
        self.directives: list[Directive] = kwargs.get("directives", [])  # type: ignore[assignment]
        self.extensions: dict[str, Any] = kwargs.get("extensions", {})  # type: ignore[assignment]

        check_directives(self.directives, location=DirectiveLocation.ARGUMENT_DEFINITION)
        self.extensions[undine_settings.CALCULATION_ARGUMENT_EXTENSIONS_KEY] = self

        self.visible_func: VisibilityFunc | None = None

    def __connect__(self, calculation: type[Calculation], name: str) -> None:
        self.calculation = calculation
        self.name = name
        self.schema_name = self.schema_name or to_schema_name(name)

        if self.description is Undefined:
            self.description = self.calculation.__attribute_docstrings__.get(name)

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(ref={self.ref!r})>"

    def __str__(self) -> str:
        arg = self.as_graphql_argument()
        return undine_settings.SDL_PRINTER.print_field_argument(self.schema_name, arg, indent=False)

    def __get__(self, instance: Calculation | None, cls: type[Calculation]) -> Any:
        if instance is None:
            return self
        return instance.__parameters__[self.name]

    def as_graphql_argument(self) -> GraphQLArgument:
        return GraphQLArgument(
            type_=self.get_field_type(),
            default_value=self.default_value,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            out_name=self.name,
            extensions=self.extensions,
        )

    def get_field_type(self) -> GraphQLInputType:
        return convert_to_graphql_type(TypeRef(self.ref), is_input=True)  # type: ignore[return-value]

    def visible(self, func: VisibilityFunc | None = None, /) -> VisibilityFunc:
        """
        Decorate a function to change the CalculationArgument's visibility in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.

        >>> class MyCalc(Calculation[int]):
        ...     value = CalculationArgument(int)
        ...
        ...     @value.visible
        ...     def value_visible(self: CalculationArgument, request: DjangoRequestProtocol) -> bool:
        ...         return False
        """
        if func is None:  # Allow `@<calculation_argument_name>.visible()`
            return self.visible  # type: ignore[return-value]
        self.visible_func = get_wrapped_func(func)
        return func

    def add_directive(self, directive: Directive, /) -> Self:
        """Add a directive to this calculation argument."""
        check_directives([directive], location=DirectiveLocation.ARGUMENT_DEFINITION)
        self.directives.append(directive)
        return self



from __future__ import annotations

import copy
from collections.abc import Hashable
from types import FunctionType
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Self, Unpack

from django.db.models import Model
from graphql import DirectiveLocation, GraphQLInputField, Undefined

from undine.converters import (
    convert_to_default_value,
    convert_to_description,
    convert_to_graphql_type,
    convert_to_input_ref,
    is_input_hidden,
    is_input_only,
    is_input_required,
    is_many,
)
from undine.dataclasses import MaybeManyOrNonNull
from undine.exceptions import MissingModelGenericError, MutationTypeKindCannotBeDeterminedError
from undine.parsers import parse_class_attribute_docstrings
from undine.query import QUERY_TYPE_REGISTRY
from undine.settings import undine_settings
from undine.typing import MutationKind, RelatedAction, TModel
from undine.utils.graphql.type_registry import (
    get_or_create_graphql_input_object_type,
    get_or_create_graphql_object_type,
)
from undine.utils.graphql.utils import check_directives
from undine.utils.model_utils import get_model_field, get_model_fields_for_graphql
from undine.utils.mutation_tree import mutate
from undine.utils.reflection import (
    FunctionEqualityWrapper,
    cache_signature_if_function,
    get_members,
    get_wrapped_func,
    is_subclass,
)
from undine.utils.text import dotpath, get_docstring, to_schema_name

if TYPE_CHECKING:
    from collections.abc import Callable, Container

    from django.db.models import QuerySet
    from graphql import GraphQLFieldResolver, GraphQLInputObjectType, GraphQLInputType, GraphQLObjectType

    from undine import QueryType
    from undine.directives import Directive
    from undine.typing import (
        ConvertionFunc,
        DefaultValueType,
        DjangoRequestProtocol,
        GQLInfo,
        InputParams,
        InputPermFunc,
        MutationTypeParams,
        ValidatorFunc,
        VisibilityFunc,
    )

__all__ = [
    "Input",
    "MutationType",
]


class MutationTypeMeta(type):
    """A metaclass that modifies how a `MutationType` is created."""

    # Set in '__new__'
    __model__: type[Model]
    __input_map__: dict[str, Input]
    __kind__: MutationKind
    __related_action__: RelatedAction
    __schema_name__: str
    __directives__: list[Directive]
    __extensions__: dict[str, Any]
    __attribute_docstrings__: dict[str, str]

    # Determined after inputs are connected
    __input_only_inputs__: dict[str, Input]
    __hidden_inputs__: dict[str, Input]
    __related_inputs__: dict[str, Input]
    __function_inputs__: dict[str, Input]
    __model_inputs__: dict[str, Input]

    def __new__(  # noqa: PLR0912, C901
        cls,
        _name: str,
        _bases: tuple[type, ...],
        _attrs: dict[str, Any],
        **kwargs: Unpack[MutationTypeParams],
    ) -> MutationTypeMeta:
        if _name == "MutationType":  # Early return for the `MutationType` class itself.
            return super().__new__(cls, _name, _bases, _attrs)

        try:
            model = MutationTypeMeta.__model__
            del MutationTypeMeta.__model__
        except AttributeError as error:
            raise MissingModelGenericError(name=_name, cls="MutationType") from error

        kind = kwargs.get("kind")
        if kind is None:
            if "create" in _name.lower():
                mutation_kind = MutationKind.create
            elif "update" in _name.lower():
                mutation_kind = MutationKind.update
            elif "delete" in _name.lower():
                mutation_kind = MutationKind.delete
            elif "__mutate__" in _attrs or "__bulk_mutate__" in _attrs:
                mutation_kind = MutationKind.custom
            else:
                raise MutationTypeKindCannotBeDeterminedError(name=_name)
        else:
            mutation_kind = MutationKind(kind)

        auto = kwargs.get("auto", undine_settings.AUTOGENERATION)
        exclude = set(kwargs.get("exclude", []))

        if auto and "pk" not in _attrs and mutation_kind.requires_pk:
            field = get_model_field(model=model, lookup="pk")
            _attrs["pk"] = Input(field, required=True)

        if auto and mutation_kind.should_use_autogeneration:
            exclude |= set(_attrs)
            if mutation_kind.no_pk:
                exclude.add("pk")
            _attrs |= get_inputs_for_model(model, exclude=exclude)

        mutation_type = super().__new__(cls, _name, _bases, _attrs)

        # Members should use `__dunder__` names to avoid name collisions with possible `Input` names.
        mutation_type.__model__ = model
        mutation_type.__input_map__ = get_members(mutation_type, Input)
        mutation_type.__kind__ = mutation_kind
        mutation_type.__related_action__ = RelatedAction(kwargs.get("related_action", RelatedAction.null))
        mutation_type.__schema_name__ = kwargs.get("schema_name", _name)
        mutation_type.__directives__ = kwargs.get("directives", [])
        mutation_type.__extensions__ = kwargs.get("extensions", {})
        mutation_type.__attribute_docstrings__ = parse_class_attribute_docstrings(mutation_type)

        check_directives(mutation_type.__directives__, location=DirectiveLocation.INPUT_OBJECT)
        mutation_type.__extensions__[undine_settings.MUTATION_TYPE_EXTENSIONS_KEY] = mutation_type

        for name, input_ in mutation_type.__input_map__.items():
            input_.__connect__(mutation_type, name)  # type: ignore[arg-type]

        def get_input_subset(ftr: Callable[[Input], bool]) -> dict[str, Input]:
            return {k: v for k, v in mutation_type.__input_map__.items() if ftr(v)}

        mutation_type.__input_only_inputs__ = get_input_subset(lambda i: i.input_only)
        mutation_type.__hidden_inputs__ = get_input_subset(lambda i: i.hidden)
        mutation_type.__related_inputs__ = get_input_subset(lambda i: is_subclass(i.ref, MutationType))
        mutation_type.__function_inputs__ = get_input_subset(lambda i: isinstance(i.ref, FunctionType))
        mutation_type.__model_inputs__ = get_input_subset(lambda i: is_subclass(i.ref, Model))

        return mutation_type

    def __str__(cls) -> str:
        return undine_settings.SDL_PRINTER.print_input_object_type(cls.__input_type__())

    def __getitem__(cls, model: type[TModel]) -> type[MutationType[TModel]]:
        # Note that this should be cleaned up in '__new__',
        # but is not if an error occurs in the class body of the defined 'MutationType'!
        MutationTypeMeta.__model__ = model
        return cls  # type: ignore[return-value]

    def __input_type__(cls) -> GraphQLInputObjectType:
        """Create the `GraphQLInputObjectType` for this `MutationType`."""
        return get_or_create_graphql_input_object_type(
            name=cls.__schema_name__,
            fields=FunctionEqualityWrapper(cls.__input_fields__, context=cls),
            description=get_docstring(cls),
            extensions=cls.__extensions__,
            out_type=cls.__convert_input__,
        )

    def __input_fields__(cls) -> dict[str, GraphQLInputField]:
        """Defer creating fields so that self-referential related inputs can be created."""
        return {
            input_.schema_name: input_.as_graphql_input_field()
            for input_ in cls.__input_map__.values()
            if not input_.hidden
        }

    def __query_type__(cls) -> type[QueryType]:
        """Get the `QueryType` for this `MutationType`."""
        return QUERY_TYPE_REGISTRY[cls.__model__]

    def __output_type__(cls) -> GraphQLObjectType:
        """Create the GraphQL `ObjectType` for this `MutationType`."""
        query_type = cls.__query_type__()

        if cls.__kind__ == MutationKind.delete:
            field = query_type.__field_map__["pk"]
            return get_or_create_graphql_object_type(
                name=cls.__schema_name__ + "Output",
                fields={"pk": field.as_graphql_field()},
            )

        return query_type.__output_type__()

    def __convert_input__(cls, input_data: dict[str, Any]) -> dict[str, Any]:
        """Enables additional conversion of input data on per-input basis."""
        for key, value in input_data.items():
            inpt = cls.__input_map__.get(key)
            if inpt is not None and inpt.convertion_func is not None:
                input_data[key] = inpt.convertion_func(inpt, value)
        return input_data

    def __add_directive__(cls, directive: Directive, /) -> Self:
        """Add a directive to this mutation."""
        check_directives([directive], location=DirectiveLocation.INPUT_OBJECT)
        cls.__directives__.append(directive)
        return cls


class MutationType(Generic[TModel], metaclass=MutationTypeMeta):
    """
    A class for creating a mutation in the GraphQL schema based on a Django Model.

    Must set the Django Model this `MutationType` is for using the generic type argument.

    The following parameters can be passed in the class definition:

    `kind: Literal["create", "update", "delete", "related", "custom"] = <inferred>`
        The kind of mutation this is. Try to infer from mutation type if not provided.

    `related_action: Literal["null", "delete", "ignore"] = "null"`
        The action to take for existing related objects that are not included in the input.

    `auto: bool = <AUTOGENERATION setting>`
        Whether to add `Input` attributes for all Model fields automatically.

    `exclude: list[str] = []`
        Model fields to exclude from automatically added `Input` attributes.

    `schema_name: str = <class name>`
        Override name for the `InputObjectType` for this `MutationType` in the GraphQL schema.

    `directives: list[Directive] = []`
        `Directives` to add to the created `InputObjectType`.

    `extensions: dict[str, Any] = {}`
        GraphQL extensions for the created `InputObjectType`.

    >>> class CreateTaskMutation(MutationType[Task]): ...
    """

    # Members should use `__dunder__` names to avoid name collisions with possible `Input` names.

    # Set in metaclass
    __model__: ClassVar[type[Model]]
    __input_map__: ClassVar[dict[str, Input]]
    __kind__: ClassVar[MutationKind]
    __related_action__: ClassVar[RelatedAction]
    __schema_name__: ClassVar[str]
    __directives__: ClassVar[list[Directive]]
    __extensions__: ClassVar[dict[str, Any]]
    __attribute_docstrings__: ClassVar[dict[str, str]]

    # Determined after inputs are connected
    __input_only_inputs__: ClassVar[dict[str, Input]]
    __hidden_inputs__: ClassVar[dict[str, Input]]
    __related_inputs__: ClassVar[dict[str, Input]]
    __function_inputs__: ClassVar[dict[str, Input]]
    __model_inputs__: ClassVar[dict[str, Input]]

    @classmethod
    def __mutate__(cls, instance: TModel, info: GQLInfo, input_data: dict[str, Any]) -> Any:
        """Method used for single object mutations."""
        return mutate(model=cls.__model__, data=input_data, related_action=cls.__related_action__)

    @classmethod
    def __bulk_mutate__(cls, instances: list[TModel], info: GQLInfo, input_data: list[dict[str, Any]]) -> Any:
        """Method used for bulk mutations."""
        return mutate(model=cls.__model__, data=input_data, related_action=cls.__related_action__)

    @classmethod
    def __permissions__(cls, instance: TModel, info: GQLInfo, input_data: dict[str, Any]) -> None:
        """Check permissions for a mutation using this `MutationType`."""

    @classmethod
    def __validate__(cls, instance: TModel, info: GQLInfo, input_data: dict[str, Any]) -> None:
        """Validate all input data given to this `MutationType`."""

    @classmethod
    def __after__(cls, instance: TModel, info: GQLInfo, input_data: dict[str, Any]) -> None:
        """A function that is run after a mutation using this `MutationType` has been executed."""

    @classmethod
    def __filter_queryset__(cls, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:
        """A function that is used to filter the queryset returned by this `MutationType`."""
        return queryset

    @classmethod
    def __is_visible__(cls, request: DjangoRequestProtocol) -> bool:
        """
        Determine if the given `MutationType` is visible in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.
        """
        return True


class Input:
    """
    A class for defining a possible input for a mutation.
    Represents an input field on a GraphQL `InputObjectType` for the `MutationType` this is added to.

    >>> class CreateTaskMutation(MutationType[Task]):
    ...     name = Input()
    """

    def __init__(self, ref: Any = None, **kwargs: Unpack[InputParams]) -> None:
        """
        Create a new Input.

        :param ref: Reference to build the input from. Must be convertable by the `convert_to_input_ref` function.
                    If not provided, use the name of the attribute this is assigned to in the `MutationType` class.
        :param many: Whether the `Input` should return a non-null list of the referenced type.
        :param required: Whether the input should be required.
        :param default_value: Value to use for the input if none is provided. Also makes the input not required,
                              if not otherwise specified. Must be a valid GraphQL default value.
        :param input_only: If `True`, the value for this `Input` is not included when the mutation is performed,
                           but is still available during permission checks and validation.
        :param hidden: If `True`, the `Input` is not included in the schema. In most cases, should also
                       add a `default_value` for the input.
        :param description: Description for the input.
        :param deprecation_reason: If the `Input` is deprecated, describes the reason for deprecation.
        :param field_name: Name of the field in the Django model. If not provided, use the name of the attribute.
        :param schema_name: Actual name of the `Input` in the GraphQL schema. Can be used to alias the `Input`
                            for the schema, or when the desired name is a Python keyword (e.g. `if` or `from`).
        :param directives: GraphQL directives for the `Input`.
        :param extensions: GraphQL extensions for the `Input`.
        """
        self.ref: Any = cache_signature_if_function(ref, depth=1)

        self.many: bool = kwargs.get("many", Undefined)  # type: ignore[assignment]
        self.required: bool = kwargs.get("required", Undefined)  # type: ignore[assignment]
        self.input_only: bool = kwargs.get("input_only", Undefined)  # type: ignore[assignment]
        self.hidden: bool = kwargs.get("hidden", Undefined)  # type: ignore[assignment]
        self.default_value: DefaultValueType = kwargs.get("default_value", Undefined)
        self.description: str | None = kwargs.get("description", Undefined)  # type: ignore[assignment]
        self.deprecation_reason: str | None = kwargs.get("deprecation_reason")
        self.field_name: str = kwargs.get("field_name", Undefined)  # type: ignore[assignment]
        self.schema_name: str = kwargs.get("schema_name", Undefined)  # type: ignore[assignment]
        self.directives: list[Directive] = kwargs.get("directives", [])
        self.extensions: dict[str, Any] = kwargs.get("extensions", {})

        check_directives(self.directives, location=DirectiveLocation.INPUT_FIELD_DEFINITION)
        self.extensions[undine_settings.INPUT_EXTENSIONS_KEY] = self

        self.validator_func: ValidatorFunc | None = None
        self.permissions_func: InputPermFunc | None = None
        self.convertion_func: ConvertionFunc | None = None
        self.visible_func: VisibilityFunc | None = None

    def __connect__(self, mutation_type: type[MutationType], name: str) -> None:
        """Connect this `Input` to the given `MutationType` using the given name."""
        self.mutation_type = mutation_type
        self.name = name
        self.field_name = self.field_name or name
        self.schema_name = self.schema_name or to_schema_name(name)

        if isinstance(self.ref, str):
            self.field_name = self.ref

        self.ref = convert_to_input_ref(self.ref, caller=self)

        if self.many is Undefined:
            self.many = is_many(self.ref, model=self.mutation_type.__model__, name=self.field_name)
        if self.input_only is Undefined:
            all_inputs_used = self.mutation_type.__kind__.all_inputs_used_by_default
            self.input_only = False if all_inputs_used else is_input_only(self.ref, caller=self)
        if self.hidden is Undefined:
            self.hidden = is_input_hidden(self.ref, caller=self)
        if self.default_value is Undefined and self.mutation_type.__kind__.should_include_default_value:
            self.default_value = convert_to_default_value(self.ref, caller=self)
        if self.required is Undefined:
            self.required = is_input_required(self.ref, caller=self)
        if self.description is Undefined:
            self.description = self.mutation_type.__attribute_docstrings__.get(name)
            if self.description is None:
                self.description = convert_to_description(self.ref)

        if not isinstance(self.default_value, Hashable):
            handle_non_hashable_default_values(self)

    def __call__(self, ref: GraphQLFieldResolver, /) -> Input:
        """Called when using as decorator with parenthesis: @Input(...)"""
        self.ref = cache_signature_if_function(ref, depth=1)
        return self

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(ref={self.ref!r})>"

    def __str__(self) -> str:
        inpt = self.as_graphql_input_field()
        return undine_settings.SDL_PRINTER.print_input_field(self.schema_name, inpt, indent=False)

    def as_graphql_input_field(self) -> GraphQLInputField:
        return GraphQLInputField(
            type_=self.get_field_type(),
            default_value=self.default_value,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            out_name=self.name,
            extensions=self.extensions,
        )

    def get_field_type(self) -> GraphQLInputType:
        value = MaybeManyOrNonNull(self.ref, many=self.many, nullable=not self.required)
        return convert_to_graphql_type(value, model=self.mutation_type.__model__, is_input=True)  # type: ignore[return-value]

    def validate(self, func: ValidatorFunc | None = None, /) -> ValidatorFunc:
        """
        Decorate a function to add validation for this Input.

        >>> class TaskCreateMutation(MutationType[Task]):
        ...     name = Input()
        ...
        ...     @name.validate
        ...     def name_validate(self: Task, info: GQLInfo, value: str) -> None:
        ...         raise GraphQLValidationError
        """
        if func is None:  # Allow `@<input_name>.validate()`
            return self.validate  # type: ignore[return-value]
        self.validator_func = get_wrapped_func(func)
        return func

    def permissions(self, func: InputPermFunc | None = None, /) -> InputPermFunc:
        """
        Decorate a function to add it as a permission check for this Input.

        >>> class TaskCreateMutation(MutationType[Task]):
        ...     name = Input()
        ...
        ...     @name.permissions
        ...     def name_permissions(self: Task, info: GQLInfo, value: str) -> None:
        ...         raise GraphQLPermissionError
        """
        if func is None:  # Allow `@<input_name>.permissions()`
            return self.permissions  # type: ignore[return-value]
        self.permissions_func = get_wrapped_func(func)
        return func

    def convert(self, func: ConvertionFunc | None = None, /) -> ConvertionFunc:
        """
        Decorate a function to add it as a convertion function for this Input.

        >>> class TaskCreateMutation(MutationType[Task]):
        ...     name = Input()
        ...
        ...     @name.convert
        ...     def name_convert(self: Input, value: str) -> str:
        ...         return value.upper()
        """
        if func is None:  # Allow `@<input_name>.convert()`
            return self.convert  # type: ignore[return-value]
        self.convertion_func = get_wrapped_func(func)
        return func

    def visible(self, func: VisibilityFunc | None = None, /) -> VisibilityFunc:
        """
        Decorate a function to change the Input's visibility in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.

        >>> class TaskCreateMutation(MutationType[Task]):
        ...     name = Input()
        ...
        ...     @name.visible
        ...     def name_visible(self: Input, request: DjangoRequestProtocol) -> bool:
        ...         return False
        """
        if func is None:  # Allow `@<input_name>.visible()`
            return self.visible  # type: ignore[return-value]
        self.visible_func = get_wrapped_func(func)
        return func

    def add_directive(self, directive: Directive, /) -> Self:
        """Add a directive to this input."""
        check_directives([directive], location=DirectiveLocation.INPUT_FIELD_DEFINITION)
        self.directives.append(directive)
        return self


def get_inputs_for_model(model: type[Model], *, exclude: Container[str] = ()) -> dict[str, Input]:
    """Add `Inputs` for all the given model's fields, except those in the 'exclude' list."""
    from django.contrib.contenttypes.fields import GenericForeignKey  # noqa: PLC0415

    result: dict[str, Input] = {}

    remove_inputs: set[str] = set()

    for model_field in get_model_fields_for_graphql(model, exclude_nonsaveable=True):
        field_name = model_field.name

        is_primary_key = bool(getattr(model_field, "primary_key", False))
        if is_primary_key:
            field_name = "pk"

        # If a 'GenericForeignKey' is added, don't include the 'content type'
        # and 'object ID' inputs, see 'GenericForeignKey' docs in Django.
        if isinstance(model_field, GenericForeignKey):
            remove_inputs.add(model_field.ct_field)
            remove_inputs.add(model_field.fk_field)

        if field_name in exclude:
            continue

        result[field_name] = Input(model_field)

    for field_name in remove_inputs:
        result.pop(field_name, None)

    return result


def handle_non_hashable_default_values(input_: Input) -> None:
    """
    If the Input's default value is not hashable (i.e. a list or a dict),
    we need to make a copy of the input value looks like it's the default value.
    Otherwise, mutations could change the default value and cause unexpected behavior.
    """
    user_func = input_.convertion_func

    def convert(inpt: Input, value: Any) -> Any:
        if value == inpt.default_value:
            value = copy.deepcopy(value)
        if user_func is not None:
            value = user_func(inpt, value)
        return value

    input_.convertion_func = convert



from __future__ import annotations

from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey
from django.db.models import Model

from undine import Input, MutationType
from undine.converters import is_input_hidden
from undine.dataclasses import LazyLambda, TypeRef
from undine.exceptions import ModelFieldDoesNotExistError
from undine.typing import ModelField
from undine.utils.model_utils import get_model_field
from undine.utils.reflection import get_signature


@is_input_hidden.register
def _(ref: ModelField, **kwargs: Any) -> bool:
    return ref.hidden


@is_input_hidden.register
def _(_: TypeRef, **kwargs: Any) -> bool:
    return False


@is_input_hidden.register
def _(_: LazyLambda, **kwargs: Any) -> bool:
    return False


@is_input_hidden.register
def _(ref: FunctionType, **kwargs: Any) -> bool:
    sig = get_signature(ref)
    # Has user input if function has three arguments: `(root: Model, info: GQLInfo, value: Any)`
    return len(sig.parameters) != 3  # noqa: PLR2004


@is_input_hidden.register
def _(_: type[MutationType], **kwargs: Any) -> bool:
    return False


@is_input_hidden.register  # Required for Django<5.1
def _(ref: GenericForeignKey, **kwargs: Any) -> bool:
    return ref.hidden


@is_input_hidden.register
def _(_: type[Model], **kwargs: Any) -> bool:
    caller: Input = kwargs["caller"]

    try:
        field = get_model_field(model=caller.mutation_type.__model__, lookup=caller.field_name)
    except ModelFieldDoesNotExistError:
        return False

    return is_input_hidden(field, **kwargs)



from __future__ import annotations

import dataclasses
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from graphql import Undefined

from undine.typing import TModel

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from types import UnionType

    from django.contrib.contenttypes.fields import GenericForeignKey
    from django.db.models import Model, OrderBy, Q, QuerySet
    from graphql import FieldNode, InlineFragmentNode

    from undine import QueryType
    from undine.pagination import PaginationHandler
    from undine.typing import (
        DispatchProtocol,
        DjangoExpression,
        LiteralArg,
        QuerySetMap,
        RelatedField,
        RelationType,
        TypeHint,
    )

__all__ = [
    "AbstractSelections",
    "BulkCreateKwargs",
    "FilterResults",
    "GraphQLHttpParams",
    "LazyGenericForeignKey",
    "LazyLambda",
    "LazyRelation",
    "LookupRef",
    "MaybeManyOrNonNull",
    "OptimizationWithPagination",
    "OrderResults",
    "Parameter",
    "RelInfo",
    "RootAndInfoParams",
    "TypeRef",
    "UnionFilterRef",
]


@dataclasses.dataclass(frozen=True, slots=True)
class Parameter:
    """Represents a parameter for a function."""

    name: str
    annotation: type | UnionType
    default_value: Any = Undefined


@dataclasses.dataclass(frozen=True, slots=True)
class FilterResults:
    """Holds the results of a QueryType filtering operation."""

    filters: list[Q]
    aliases: dict[str, DjangoExpression]
    distinct: bool
    none: bool = False
    filter_count: int = 0


@dataclasses.dataclass(frozen=True, slots=True)
class OrderResults:
    """Holds the results of a QueryType ordering operation."""

    order_by: list[OrderBy]
    aliases: dict[str, DjangoExpression]
    order_count: int = 0


@dataclasses.dataclass(frozen=True, slots=True)
class GraphQLHttpParams:
    """Holds the parameters from a GraphQL HTTP request."""

    document: str
    variables: dict[str, Any]
    operation_name: str | None
    extensions: dict[str, Any]


@dataclasses.dataclass(frozen=True, slots=True)
class TypeRef:
    """A reference to a type used by converters."""

    value: TypeHint

    total: bool = True
    """If the type hint is in a TypedDict, whether the TypedDict has totality of not."""


@dataclasses.dataclass(frozen=True, slots=True)
class MaybeManyOrNonNull:
    """A reference to a maybe null or many type used by converters."""

    value: Any
    many: bool
    nullable: bool


@dataclasses.dataclass(frozen=True, slots=True)
class LookupRef:
    """A reference to a lookup expression used by converters."""

    ref: Any
    lookup: str


@dataclasses.dataclass(frozen=True, slots=True)
class UnionFilterRef:
    """
    A string reference that references a model field
    that exists in all models of a FilterSet used for a UnionType.
    """

    ref: str
    models: tuple[type[Model], ...]


@dataclasses.dataclass(frozen=True, slots=True)
class ValidatedPaginationArgs:
    """Pagination arguments that have been validated."""

    after: int | None
    before: int | None
    first: int | None
    last: int | None


@dataclasses.dataclass(slots=True)
class OptimizationWithPagination(Generic[TModel]):
    """Pagination arguments that have been validated."""

    queryset: QuerySet[TModel]
    pagination: PaginationHandler


@dataclasses.dataclass(slots=True)
class QuerySetMapWithPagination(Generic[TModel]):
    """Pagination arguments that have been validated."""

    queryset_map: QuerySetMap
    pagination: PaginationHandler


@dataclasses.dataclass(frozen=True, slots=True)
class RootAndInfoParams:
    root_param: str | None
    info_param: str | None


@dataclasses.dataclass(frozen=True, slots=True)
class LazyRelation:
    """Represents a lazily evaluated field for a related field."""

    field: RelatedField

    def get_type(self) -> type[QueryType]:
        from undine.query import QUERY_TYPE_REGISTRY  # noqa: PLC0415

        return QUERY_TYPE_REGISTRY[self.field.related_model]  # type: ignore[index]


@dataclasses.dataclass(frozen=True, slots=True)
class LazyGenericForeignKey:
    """Represents a lazily evaluated Field for a generic foreign key."""

    field: GenericForeignKey

    def get_types(self) -> list[type[QueryType]]:
        from undine.query import QUERY_TYPE_REGISTRY  # noqa: PLC0415
        from undine.utils.model_utils import generic_relations_for_generic_foreign_key  # noqa: PLC0415

        return [
            QUERY_TYPE_REGISTRY[field.remote_field.related_model]  # type: ignore[index]
            for field in generic_relations_for_generic_foreign_key(self.field)
            if field.remote_field.related_model in QUERY_TYPE_REGISTRY  # type: ignore[index]
        ]


@dataclasses.dataclass(frozen=True, slots=True)
class LazyLambda:
    """Represents a lazily evaluated object behind a lambda function."""

    callback: Callable[[], type[QueryType]]


@dataclasses.dataclass(frozen=True, slots=True)
class RelInfo:
    """Holds information about a related field on a model."""

    relation_type: RelationType

    # Relation source
    field_name: str
    model: type[Model]
    model_pk_type: type
    nullable: bool

    # Relation target
    related_name: str | None  # Null for generic foreign keys
    related_model: type[Model] | None  # Null for generic foreign keys
    related_model_pk_type: type
    related_nullable: bool


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class BulkCreateKwargs(Mapping[str, Any]):
    """Arguments to use in bulk create."""

    update_fields: set[str] | None = dataclasses.field(default=None)

    update_conflicts: bool = dataclasses.field(init=False, default=False)
    unique_fields: set[str] | None = dataclasses.field(init=False, default=None)

    def __post_init__(self) -> None:
        if self.update_fields:
            object.__setattr__(self, "update_conflicts", True)
            object.__setattr__(self, "unique_fields", {"pk"})

    def __iter__(self) -> Iterator[str]:
        return iter(dataclasses.asdict(self))

    def __len__(self) -> int:
        return len(dataclasses.asdict(self))

    def __getitem__(self, key: str) -> Any:
        return dataclasses.asdict(self)[key]

    def __bool__(self) -> bool:
        return self.update_conflicts and bool(self.update_fields) and bool(self.unique_fields)


T = TypeVar("T")


@dataclasses.dataclass(frozen=True, slots=True)
class DispatchImplementations(Generic[T]):
    """Holds the implementations of a `FunctionDispatcher`."""

    types: dict[type, DispatchProtocol[T]] = dataclasses.field(default_factory=dict)
    instances: dict[object, DispatchProtocol[T]] = dataclasses.field(default_factory=dict)
    literals: dict[LiteralArg, DispatchProtocol[T]] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass(slots=True)
class AbstractSelections:
    """Flattened selections for an abstract type."""

    field_nodes: list[FieldNode] = dataclasses.field(default_factory=list)
    inline_fragments: list[InlineFragmentNode] = dataclasses.field(default_factory=list)
